import ocs

import txaio
txaio.use_twisted()

from twisted.application.internet import backoffPolicy
from twisted.internet import reactor, task, threads
from twisted.internet.defer import inlineCallbacks, Deferred, DeferredList, FirstError, maybeDeferred
from twisted.internet.error import ReactorNotRunning

from twisted.python import log
from twisted.logger import formatEvent, FileLogObserver

from autobahn.wamp.types import ComponentConfig, SubscribeOptions
from autobahn.twisted.wamp import ApplicationSession, ApplicationRunner
from autobahn.twisted.util import sleep as dsleep
from autobahn.wamp.exception import ApplicationError, TransportLost
from autobahn.exception import Disconnected
from .ocs_twisted import in_reactor_context

import json
import math
import time
import datetime
import signal
import socket
import os
from ocs import client_t
from ocs import ocs_feed
from ocs.base import OpCode


def init_site_agent(args, address=None):
    """
    Create ApplicationSession and ApplicationRunner instances, set up
    to communicate on the chosen WAMP realm.

    Args:
        args (argparse.Namespace): The arguments, as processed by
            ocs.site_config.

    Returns: (agent, runner).
    """
    if address is None:
        address = '%s.%s' % (args.address_root, args.instance_id)
    server, realm = args.site_hub, args.site_realm
    # txaio.start_logging(level='debug')
    agent = OCSAgent(ComponentConfig(realm, {}), args, address=address,
                     class_name=getattr(args, 'agent_class', None))
    runner = ApplicationRunner(server, realm)
    return agent, runner


def log_formatter(event):
    text = formatEvent(event)
    t = datetime.datetime.utcnow()
    date_str = t.strftime("%Y-%m-%dT%H-%M-%S.%f")
    return '%s %s\n' % (date_str, text)


class OCSAgent(ApplicationSession):
    """OCSAgent is used to connect blocking device control code to the
    OCS.  OCSAgent is an ApplicationSession and its methods are all
    run in the twisted main Reactor thread.

    To make use of OCSAgent, the user instantiates it (perhaps through
    init_ocs_agent) with a particular agent_address.  Then the user
    registers task and process functions by calling register_task()
    and register_process().  These are (blocking) functions that will
    be called, in their own twisted thread, when the "start" method is
    request.

    The OCSAgent automatically registers handlers in WAMP, namely:

      {agent_address} - the management_handler function, which
        responds to queries about what tasks and processes are exposed
        by this agent.

      {agent_address}.ops - the device_handler function, which accepts
        Operation commands (start, status, etc.) for all Tasks and
        Processes.

    The OCSAgent also makes use of pubsub channels:

      {agent_address}.feed - a channel to which any session status
        updates are published (written by the Agent; subscribed by any
        interested Control Tools).

    """

    def __init__(self, config, site_args, address=None, class_name=None):
        ApplicationSession.__init__(self, config)
        self.log.info("Using OCS version {v}", v=ocs.__version__)
        self.site_args = site_args
        self.tasks = {}       # by op_name
        self.processes = {}   # by op_name
        self.feeds = {}
        self.sessions = {}    # by op_name, single OpSession.
        self.next_session_id = 0
        self.session_archive = {}  # by op_name, lists of OpSession.
        self.agent_address = address
        self.class_name = class_name
        self.registered = False
        self.log = txaio.make_logger()
        self.heartbeat_call = None
        self._heartbeat_on = True
        self.agent_session_id = str(time.time())
        self.startup_ops = []  # list of (op_type, op_name, op_params)
        self.startup_subs = []  # list of dicts with params for subscribe call
        self.subscribed_topics = set()
        self.subscriptions = []  # autobahn.wamp.request.Subscription objects
        self.realm_joined = False
        self.first_time_startup = True

        # Attach the logger.
        log_dir, log_file = site_args.log_dir, None
        if log_dir is not None:
            if not log_dir.startswith('/'):
                if site_args.working_dir is None:
                    self.log.error('Cannot use relative log_dir without '
                                   'explicit working_dir.')
                else:
                    log_dir = os.path.join(site_args.working_dir, log_dir)
            if log_dir is not None and os.path.exists(log_dir):
                log_file = '%s/%s.log' % (log_dir, self.agent_address)
                try:
                    fout = open(log_file, 'a')
                    log.addObserver(FileLogObserver(fout, log_formatter))
                except PermissionError:
                    self.log.error('Permissions error writing to log file %s' % log_file)
            else:
                self.log.error('Log directory does not exist: %s' % log_dir)

        # Can we log already?
        self.log.info('ocs: starting %s @ %s' % (str(self.__class__), address))
        self.log.info('log_file is apparently %s' % (log_file))

    @inlineCallbacks
    def _stop_all_running_sessions(self):
        """Stops all currently running sessions."""
        self.log.info('Stopping all running sessions')
        for session in self.sessions:
            if self.sessions[session] is not None:
                self.log.info("Stopping session {sess}", sess=session)
                self.log.debug("session details: {sess}",
                               sess=self.sessions[session].encoded())
                # Only try to stop starting or running sessions
                if self.sessions[session].status not in ['stopping', 'done']:
                    if session in self.tasks:
                        yield self.abort(session)
                    elif session in self.processes:
                        yield self.stop(session)
        # Give a second for processes to stop cleanly
        yield dsleep(3)

    """
    Methods below are implementations of the ApplicationSession.
    """

    def onConnect(self):
        # Define signal handlers
        @inlineCallbacks
        def signal_handler(sig, frame):
            self.log.info('caught {signal}!', signal=signal.Signals(sig).name)
            yield self._shutdown()

        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)

        self.log.info('transport connected')
        self.join(self.config.realm)

    def onChallenge(self, challenge):
        self.log.info('authentication challenge received')

    def _store_subscription(self, subscription, *args, **kwargs):
        self.subscriptions.append(subscription)

    def _unsub_error(self, *args, **kwargs):
        # This just swallows an wamp.error.no_such_subscription exception
        # It is always generated when unsubscribing from stale subscriptions
        pass

    def _unsubscribe_all(self):
        for sub in self.subscriptions:
            self.log.debug('Unsubscribing {sub}', sub=sub)
            try:
                d = sub.unsubscribe()
                d.addErrback(self._unsub_error)
            except Exception as e:
                self.log.error('Error encountered when unsubscribing {sub}:', sub=sub)
                self.log.error('{error}', error=e)

        self.subscriptions = []
        self.subscribed_topics = set()

    @inlineCallbacks
    def onJoin(self, details):
        self.log.info('session joined: {x}', x=details)
        # Get an address somehow...
        if self.agent_address is None:
            self.agent_address = 'observatory.random'
        # Register our processes...
        # Register the device interface functions.
        try:
            yield self.register(self._ops_handler, self.agent_address + '.ops')
            yield self.register(self._management_handler, self.agent_address)
        except ApplicationError as e:
            self.log.error('Failed to register basic handlers!  '
                           'Error: {error}', error=e)
            if e.error == 'wamp.error.not_authorized':
                self.log.error('Are the WAMP realm and OCS address_root consistent '
                               'in OCS site config and crossbar config.json?')
            elif e.error == 'wamp.error.procedure_already_exists':
                self.log.error('Is this agent already running? '
                               'agent_address="{agent_address}"',
                               agent_address=self.agent_address)
            self.leave()
            return

        self.register_feed("heartbeat", max_messages=1)

        def heartbeat():
            if self._heartbeat_on:
                self.log.debug(' {:.1f} {address} heartbeat '
                               .format(time.time(), address=self.agent_address))

                op_codes = {}
                for name, session in self.sessions.items():
                    if session is None:
                        op_codes[name] = OpCode.NONE.value
                    else:
                        op_codes[name] = session.op_code.value
                self.publish_to_feed("heartbeat", op_codes, from_reactor=True)

        self.heartbeat_call = task.LoopingCall(heartbeat)
        self.heartbeat_call.start(1.0)  # Calls the hearbeat every second

        # Remove old subscriptions
        self._unsubscribe_all()

        # Subscribe to startup_subs
        def _subscribe_fail(*args, **kwargs):
            self.log.error('Failed to subscribe to a feed or feed pattern; possible configuration problem.')
            self.log.error(str(args) + str(kwargs))
            self.leave()

        for sub in self.startup_subs:
            d = maybeDeferred(self.subscribe, **sub)
            d.addCallback(self._store_subscription)
            d.addErrback(_subscribe_fail)

        # Now do the startup activities, only the first time we join
        if self.first_time_startup:
            for op_type, op_name, op_params in self.startup_ops:
                self.log.info('startup-op: launching %s' % op_name)
                if op_params is True:
                    op_params = {}
                self.start(op_name, op_params)
            self.first_time_startup = False

        self.realm_joined = True

    @inlineCallbacks
    def onLeave(self, details):
        self.log.info('session left: {}'.format(details))
        if self.heartbeat_call is not None:
            self.heartbeat_call.stop()

        # Normal shutdown
        if details.reason == "wamp.close.normal":
            yield self._stop_all_running_sessions()

        self.disconnect()

        # Unsub from all topics, since we've left the realm
        self.subscribed_topics = set()
        self.realm_joined = False

    @inlineCallbacks
    def _shutdown(self):
        # Stop all sessions and then stop the reactor
        yield self._stop_all_running_sessions()
        try:
            self.log.info('stopping reactor')
            reactor.stop()
        except ReactorNotRunning:
            pass

    @inlineCallbacks
    def onDisconnect(self):
        self.log.info('transport disconnected')
        self.log.info('waiting for reconnection')

        # Wait to see if we reconnect before stopping the reactor
        timeout = self.site_args.crossbar_timeout

        # Wait forever
        if timeout == 0:
            return

        # compute_delay(attempts): Delay in seconds given number of attempts
        # Twisted has an exponential backoff interval that prevents flooding
        # reconnect attempts. We follow that roughly for checking the Agent has
        # reconnected to crossbar up to a 30 second delay.
        compute_delay = backoffPolicy(maxDelay=30.0)

        # Disconnect after timeout
        disconnected_at = time.time()
        attempt = 0
        while time.time() - disconnected_at < timeout:
            attempt += 1  # twisted also starts at 1 attempt

            # successful reconnection
            if self.realm_joined:
                return

            time_left = timeout - (time.time() - disconnected_at)
            self.log.info('waiting at least {} more seconds before giving up'.format(time_left))
            delay = compute_delay(attempt)
            yield dsleep(delay)

        # Shutdown after timeout expires
        yield self._shutdown()

    """The methods below provide OCS framework support."""

    def encoded(self):
        """
        Returns a dict describing this Agent.  Includes 'agent_address',
        and lists of 'feeds', 'tasks', and 'processes'.
        """
        return {
            'agent_address': self.agent_address,
            'session_id': self.agent_session_id,
            'feeds': [f[1].encoded() for f in self.feeds.items()],
            'tasks': list(self.tasks.keys()),
            'processes': list(self.processes.keys())
        }

    def _ops_handler(self, action, op_name, params=None, timeout=None):
        if action == 'start':
            return self.start(op_name, params=params)
        if action == 'stop':
            return self.stop(op_name, params=params)
        if action == 'abort':
            return self.abort(op_name, params=params)
        if action == 'wait':
            return self.wait(op_name, timeout=timeout)
        if action == 'status':
            return self.status(op_name)
        return (ocs.ERROR, 'No implementation for "%s"' % op_name, {})

    def _gather_sessions(self, parent):
        """Gather the session data for self.tasks or self.sessions, for return
        through the management_handler.

        Args:
          parent: either self.tasks or self.processes.

        Returns:

          A list of Operation description tuples, one per registered
          Task or Process.  Each tuple consists of elements `(name,
          session, op_info)`:

          - `name`: The name of the operation.
          - `session`: dict with OpSession.encode(() info for the
            active or most recent session.  If no such session exists
            the result will have member 'status' set to 'no_history'.
          - `op_info`: information registered about the operation,
            such as `op_type`, `docstring` and `blocking`.

        """
        result = []
        for name, op_info in sorted(parent.items()):
            session = self.sessions.get(name)
            if session is None:
                session = {'op_name': name, 'status': 'no_history'}
            else:
                session = session.encoded()
            result.append((name, session, op_info.encoded()))
        return result

    def _management_handler(self, q, **kwargs):
        """Get a description of this Agent's API.  This is for adaptive
        clients (such as MatchedClient) to construct their interfaces.

        Parameters
        ----------
        q : string
          One of 'get_api', 'get_tasks', 'get_processes', 'get_feeds',
          'get_agent_class'.

        Returns
        -------
        api_description : dict
          If the argument is 'get_api', then a dict with the following
          entries is returned:

          - 'agent_class': The class name of this agent.
          - 'instance_hostname': The host name where the Agent is
            running, as returned by socket.gethostname().
          - 'instance_pid': The PID of the Agent interpreter session,
            as returned by os.getpid().
          - 'feeds': The list of encoded feed information, tuples
            (feed_name, feed_info).
          - 'processes': The list of Process api description info, as
            returned by :func:`_gather_sessions`.
          - 'tasks': The list of Task api description info, as
            returned by :func:`_gather_sessions`.

          Passing get_X will, for some values of X, return only that
          subset of the full API; treat that as deprecated.

        """
        if q == 'get_api':
            return {
                'agent_class': self.class_name,
                'instance_hostname': socket.gethostname(),
                'instance_pid': os.getpid(),
                'feeds': [(k, v.encoded()) for k, v in self.feeds.items()],
                'processes': self._gather_sessions(self.processes),
                'tasks': self._gather_sessions(self.tasks),
            }
        if q == 'get_tasks':
            return self._gather_sessions(self.tasks)
        if q == 'get_processes':
            return self._gather_sessions(self.processes)
        if q == 'get_feeds':
            return [(k, v.encoded()) for k, v in self.feeds.items()]
        if q == 'get_agent_class':
            return self.class_name

    def register_task(self, name, func, aborter=None, blocking=True,
                      aborter_blocking=None, startup=False):
        """Register a Task for this agent.

        Args:
            name (string): The name of the Task.
            func (callable): The function that will be called to
                handle the "start" operation of the Task.
            aborter (callable): The function that will be called to
                handle the "abort" operation of the Task (optional).
            blocking (bool): Indicates that ``func`` should be
               launched in a worker thread, rather than running in the
               main reactor thread.
            aborter_blocking(bool or None): Indicates that ``aborter``
               should be run in a worker thread, rather than running
               in the main reactor thread.  Defaults to value of
               ``blocking``.
            startup (bool or dict): Controls if and how the Operation
                is launched when the Agent successfully starts up and
                connects to the WAMP realm.  If False, the Operation
                does not auto-start.  Otherwise, the Operation is
                launched on startup.  If the ``startup`` argument is a
                dictionary, this is passed to the Operation's start
                function.

        Notes:

            The functions func and aborter will be called with
            arguments (session, params) where session is the active
            OpSession and params is passed from the client.

            (Passing params to the aborter might not be supported in
            the client library so don't count on that being useful.)

        """
        self.tasks[name] = AgentTask(
            func, blocking=blocking, aborter=aborter,
            aborter_blocking=aborter_blocking)
        self.sessions[name] = None
        if startup is not False:
            self.startup_ops.append(('task', name, startup))

    def register_process(self, name, start_func, stop_func, blocking=True,
                         stopper_blocking=None, startup=False):
        """Register a Process for this agent.

        Args:
            name (string): The name of the Process.
            start_func (callable): The function that will be called to
                handle the "start" operation of the Process.
            stop_func (callable): The function that will be called to
                handle the "stop" operation of the Process.
            blocking (bool): Indicates that ``start_func`` should be
                launched in a worker thread, rather than running in
                the reactor.
            stopper_blocking (bool or None): Indicates that
                ``stop_func`` should be launched in a worker thread,
                rather than running in the reactor.  Defaults to the
                value of ``blocking``.
            startup (bool or dict): Controls if and how the Operation
                is launched when the Agent successfully starts up and
                connects to the WAMP realm.  If False, the Operation
                does not auto-start.  Otherwise, the Operation is
                launched on startup.  If the ``startup`` argument is a
                dictionary, this is passed to the Operation's start
                function.

        Notes:
            The functions start_func and stop_func will be called with
            arguments (session, params) where session is the active
            OpSession and params is passed from the client.

            (Passing params to the stop_func might not be supported in
            the client library so don't count on that being useful.)

        """
        self.processes[name] = AgentProcess(
            start_func, stop_func, blocking=blocking,
            stopper_blocking=stopper_blocking)
        self.sessions[name] = None
        if startup is not False:
            self.startup_ops.append(('process', name, startup))

    @inlineCallbacks
    def call_op(self, agent_address, op_name, action, params=None, timeout=None):
        """
        Calls ocs_agent operation.

        Args:
            agent_address (string):
                Address of the agent who registered operation
            op_name (string):
                Name of the operation
            action (string):
                Action of operation. start, stop , wait, etc.
            params (dict):
                Params passed to operation
            timeout (float):
                timeout for operation
        """
        if not in_reactor_context():
            x = yield reactor.callFromThread(
                self.call_op, agent_address, op_name, action, params=params, timeout=timeout
            )
            return x

        op = client_t.OperationClient(self, agent_address, op_name)
        try:
            x = yield op.request(action, params=params, timeout=timeout)
            return x
        except ApplicationError as e:
            self.log.warn(e.error)
            if e.error == u'wamp.error.no_such_procedure':
                self.log.warn("Operation {}.ops.{} has not been registered"
                              .format(agent_address, op_name))
            else:
                self.log.warn(e.error)
            return False

    def register_feed(self, feed_name, **kwargs):
        """
        Initializes a new feed with name ``feed_name``.

        Args:
            feed_name (string):
                name of the feed
            record (bool, optional):
                Determines if feed should be aggregated. At the moment, each agent
                can have at most one aggregated feed. Defaults to False
            agg_params (dict, optional):
                Parameters used by the aggregator and influx publisher.  See
                the ``ocs.ocs_feed.Feed`` docstring for the full list of
                aggregator params.
            buffer_time (int, optional):
                Specifies time that messages should be buffered in seconds.
                If 0, message will be published immediately.
                Defaults to 0.
            max_messages (int, optional):
                Max number of messages stored. Defaults to 20.

        Returns:
            The Feed object (which is also cached in self.feeds).
        """
        self.feeds[feed_name] = ocs_feed.Feed(self, feed_name, **kwargs)
        return self.feeds[feed_name]

    def publish_to_feed(self, feed_name, message, from_reactor=None):
        """Publish data to named feed.

        Args:
          feed_name (str): should match the name of a registered feed.
          message (serializable): data to publish.  Acceptable format
            depends on feed configuration; see Feed.publish_message.
          from_reactor (bool or None): This is deprecated; the code
            will check whether you're in a thread or not.

        Notes:
          If an unknown feed_name is passed in, an error is printed to
          the log and that's all.

          If you are running a "blocking" operation, in a thread, then
          it is best if the message is not a persistent data structure
          from your thread (especially something you might modify soon
          after this call).  The code will take a copy of your
          structure and pass that to the reactor thread, but the copy
          may not be deep enough!

        """
        if feed_name not in self.feeds:
            self.log.error("Feed {} is not registered.".format(feed_name))
            return
        # We expect that publish_message will check threading context
        # and do the right thing (as of this writing, it does).
        self.feeds[feed_name].publish_message(message)

    def subscribe(self, handler, topic, options=None, force_subscribe=False):
        """
        Subscribes to a topic for receiving events.
        Identical to ApplicationSession subscribe, but by default prevents
        re-subscription to the same topic multiple times unless
        force_subscribe=True.

        For full documentation see:
        https://autobahn.readthedocs.io/en/latest/reference/autobahn.wamp.html#autobahn.wamp.interfaces.ISession.subscribe

        Args:
            handler (callable):
                handler called with message data
            topic (string):
                uri of topic to subscribe to
            options (dict):
                Dict of subscribe options.
                To set prefix or wildcard matching, set `match` to `prefix`
                or `wildcard` respectively.
                For more info, see https://autobahn.readthedocs.io/en/latest/reference/autobahn.wamp.html#autobahn.wamp.types.SubscribeOptions
            force_subscribe (bool):
                If true, force resubscribe to an already susbscribed topic.
                Defaults to False.
        """
        if (topic not in self.subscribed_topics) or force_subscribe:
            if options is not None:
                options = SubscribeOptions(**options)
            self.subscribed_topics.add(topic)
            return super().subscribe(handler, topic=topic,
                                     options=options)
        else:
            self.log.warn("Topic {} is already subscribed.".format(topic))
            return False

    def subscribe_to_feed(self, agent_addr, feed_name, handler, options=None, force_subscribe=False):
        """
        Constructs topic feed from agent address and feedname, and subscribes to it.

        Args:
            agent_addr (str):
                Full agent address, e.g. `observatory.LS12345`
            feed_name (str):
                Feed name, e.g. `temperatures`
            handler (callable):
                handler called with message data
            options (dict):
                Dict or subscribe options. See https://autobahn.readthedocs.io/en/latest/reference/autobahn.wamp.html#autobahn.wamp.types.SubscribeOptions
            force_subscribe (bool):
                If true, force resubscribe to an already susbscribed topic.
                Defaults to False.
        """
        topic = "{}.feeds.{}".format(agent_addr, feed_name)
        return self.subscribe(handler, topic, options=options, force_subscribe=force_subscribe)

    def subscribe_on_start(self, handler, topic, options=None, force_subscribe=None):
        """
        Schedules a topic to be subscribed to OnJoin.
        See OCSAgent.subscribe's docstring.
        """
        self.startup_subs.append({
            'handler': handler,
            'topic': topic,
            'options': options,
            'force_subscribe': force_subscribe
        })

    def _handle_task_return_val(self, *args, **kw):
        try:
            (ok, message), session = args
            session.add_message(message)
            session.success = ok
            session.set_status('done')
        except BaseException:
            print('Failed to decode _handle_task_return_val args:',
                  args, kw)
            raise

    def _handle_task_error(self, *args, **kw):
        try:
            ex, session = args
            if ex.check(ParamError):
                message = 'ERROR: {}'.format(ex.getErrorMessage())
            else:
                message = 'CRASH: %s' % str(ex)
            session.add_message(message)
            session.success = False
            session.set_status('done')
        except BaseException:
            print('Failure to decode _handle_task_error args:',
                  args, kw)
            raise

    """ The functions below define the OCS API for client control of an
    Agent's Operations.  Some methods are valid on Processs, some on
    Tasks, and some on both."""

    def start(self, op_name, params=None):
        """
        Launch an operation.  Note that successful return of this function
        does not mean that the operation is running; it only means
        that the system has requested the operation to run.

        Returns tuple (status, message, session).

        Possible values for status:

          ocs.ERROR: the specified op_name is not known, or the op is
            already running (has an active session).

          ocs.OK: the Operation start routine has been launched.
        """
        print('start called for {}'.format(op_name))
        is_task = op_name in self.tasks
        is_proc = op_name in self.processes
        if is_task or is_proc:
            # Confirm it is currently idle.
            session = self.sessions.get(op_name)
            if session is not None:
                if session.status == 'done':
                    # Move to history...
                    # ...
                    # Clear from active.
                    self.sessions[op_name] = None
                else:
                    return (ocs.ERROR, 'Operation "%s" already in progress.' % op_name,
                            session.encoded())

            # Get the task/process launch function
            if is_task:
                op = self.tasks[op_name]
                msg = 'Started task "%s".' % op_name
            else:
                op = self.processes[op_name]
                msg = 'Started process "%s".' % op_name

            # Pre-process params?
            if hasattr(op.launcher, '_ocs_prescreen'):
                try:
                    handler = ParamHandler(params)
                    params = handler.batch(op.launcher._ocs_prescreen)
                except ParamError as err:
                    self.log.error("Caught ParamError during start call: {err}", err=err)
                    return (ocs.ERROR, err.msg, {})
                except Exception as err:
                    self.log.error("Caught Exception during start call: {err}", err=err)
                    return (ocs.ERROR, f'CRASH: during param pre-processing: {str(err)}', {})

            # Mark as started.
            session = OpSession(self.next_session_id, op_name, app=self)
            self.next_session_id += 1
            self.sessions[op_name] = session

            # Schedule op to run (in worker thread or reactor)
            session.d = op.launch_deferred(session, params)
            session.d.addCallback(self._handle_task_return_val, session)
            session.d.addErrback(self._handle_task_error, session)

            return (ocs.OK, msg, session.encoded())

        else:
            self.log.warn("No task called {}".format(op_name))
            return (ocs.ERROR, 'No task or process called "%s"' % op_name, {})

    @inlineCallbacks
    def wait(self, op_name, timeout=None):
        """Wait for the specified Operation to become idle, or for timeout
        seconds to elapse.  If timeout==None, the timeout is disabled
        and the function will not return until the Operation
        terminates.  If timeout<=0, then the function will return
        immediately.

        Returns (status, message, session).

        Possible values for status:

          ocs.TIMEOUT: the timeout expired before the Operation became
            idle.

          ocs.ERROR: the specified op_name is not known.

          ocs.OK: the Operation has become idle.

        """
        if not (op_name in self.tasks or op_name in self.processes):
            return (ocs.ERROR, 'Unknown operation "%s".' % op_name, {})

        session = self.sessions[op_name]
        if session is None:
            return (ocs.OK, 'Idle.', {})

        # Note that you can't just trust session.d.called to see if
        # the Op has ended.  For a "non-blocking" Operation
        # implementation (launched with task.deferLater and runs in
        # the reactor), the Deferred in session.d fires its first
        # callback, and sets called=True, when the start() function is
        # initiated, not when it completes.  Unfortunately this means
        # we have to trust session.status ... but that should be fine.
        done = False

        if session.status == 'done' or timeout is None:
            # Op is either done or we're happy to wait for it
            yield session.d
            done = True
        elif timeout < 0:
            # Op is running, but don't wait.
            pass
        else:
            # Op is running, wait for a limited duration.
            td = Deferred()
            reactor.callLater(timeout, td.callback, None)
            dl = DeferredList([session.d, td], fireOnOneCallback=True,
                              fireOnOneErrback=True, consumeErrors=True)
            try:
                yield dl
            except FirstError as e:
                assert e.index == 0  # i.e. session.d raised an error.
                td.cancel()
                e.subFailure.raiseException()
            else:
                done = (session.status == 'done')

        if done:
            success_str = {True: 'SUCCEEDED'}.get(session.success, 'FAILED')
            return (ocs.OK, f'Operation "{op_name}" is currently not running '
                    + f'({success_str}).', session.encoded())
        else:
            return (ocs.TIMEOUT, 'Operation "%s" still running; wait timed out.' % op_name,
                    session.encoded())

    def _stop_helper(self, stop_type, op_name, params):
        """Common stopper/aborter code for Process stop and Task
        abort.

        Args:
          stop_type (str): either 'stop' or 'abort'.
          op_name (str): the op_name.
          params (dict or None): Params to be passed to stopper
            function.

        """
        print(f'{stop_type} called for {op_name}')

        # Find the op and populate op_type, op, stopper, stopper_blocking.
        if op_name in self.tasks:
            op_type = 'task'
            op = self.tasks[op_name]
            stopper = op.aborter
            stopper_blocking = op.aborter_blocking
        elif op_name in self.processes:
            op_type = 'process'
            op = self.processes[op_name]
            stopper = op.stopper
            stopper_blocking = op.stopper_blocking
        else:
            return (ocs.ERROR, 'No operation called "%s".' % op_name, {})

        # Make sure the API function matches the op_type ...
        if (stop_type == 'stop' and op_type == 'task') or \
           (stop_type == 'abort' and op_type == 'process'):
            return (ocs.ERROR, f'Cannot "{stop_type}" "{op_name}" because '
                    'it is a "{op_type}".', {})

        session = self.sessions.get(op_name)
        if session is None:
            return (ocs.ERROR, 'No session active.', {})

        if session.status in ['stopping', 'done']:
            return (ocs.ERROR, f'The operation is already {session.status}', {})

        # Use callback/errback to print message to logs.
        def _callback(*args, **kw):
            try:
                ok, msg = args
            except BaseException:
                ok, msg = True, str(args)
            print(f'Stopper for "{op_name}" terminated with ok={ok} and '
                  f'message {msg}')

        def _errback(*args, **kw):
            print(f'Error calling stopper for "{op_name}"; args:',
                  args, kw)

        if stopper_blocking:
            # Launch the code in a thread.
            d2 = threads.deferToThread(stopper, session, params)
            d2.addCallback(_callback).addErrback(_errback)
        else:
            # Assume the stopper returns a Deferred (and will soon run
            # in the reactor).
            d2 = stopper(session, params)
            if not isinstance(d2, Deferred):
                # Warn but let it slide.  in the past the default was
                # to run the stopper in a worker thread.  Most
                # stoppers run very quickly so it is probably not
                # going to break much to have them run in the reactor.
                # Change this to an error after all Agents have been
                # updated for a while.
                print(f'WARNING: {op_type} {op_name} did not return '
                      'a Deferred. If the stopper is meant to be run '
                      'in a worker thread, the op should be registered '
                      'with stopper_blocking=True (Process) or '
                      'aborter_blocking=True (Task).  If the stopper '
                      'can safely run in the reactor, it should be '
                      'modified to return a Deferred.')
            else:
                d2.addCallback(_callback).addErrback(_errback)

        return (ocs.OK, f'Requested {stop_type} on {op_type} {op_name}".',
                session.encoded())

    def stop(self, op_name, params=None):
        """
        Initiate a Process stop routine.

        Returns (status, message, session).

        Possible values for status:

          ocs.ERROR: the specified op_name is not known, or refers to
            a Task.  Also returned if Process is known but not running.

          ocs.OK: the Process stop routine has been launched.
        """
        return self._stop_helper('stop', op_name, params)

    def abort(self, op_name, params=None):
        """
        Initiate a Task abort routine.

        Returns (status, message, session).

        Possible values for status:

          ocs.ERROR: the specified op_name is not known, or refers to
            a Process.  Also returned if Task is known but not running.

          ocs.OK: the Process stop routine has been launched.

        """
        return self._stop_helper('abort', op_name, params)

    def status(self, op_name, params=None):
        """
        Get an Operation's session data.

        Returns (status, message, session).  When there is no session
        data available, an empty dictionary is returned instead.

        Possible values for status:

          ocs.ERROR: the specified op_name is not known.

          ocs.OK: the op_name was recognized.
        """
        if op_name in self.tasks or op_name in self.processes:
            session = self.sessions.get(op_name)
            if session is None:
                return (ocs.OK, 'No session active.', {})
            else:
                return (ocs.OK, 'Session active.', session.encoded())
        else:
            return (ocs.ERROR, 'No task or process called "%s"' % op_name, {})


class AgentOp:
    def launch_deferred(self, session, params):
        """Launch the operation using the launcher function, either in
        a worker thread (self.blocking) or in the reactor (not
        self.blocking).  Return a Deferred.  Prior to executing the
        operation code, set session state to "running".

        """
        def _running_wrapper(session, params):
            session.set_status('running')
            return self.launcher(session, params)

        if self.blocking:
            # Launch, soon, in a blockable worker thread.
            return threads.deferToThread(_running_wrapper, session, params)
        else:
            # Launch, soon, in the main reactor thread.
            return task.deferLater(reactor, 0, _running_wrapper, session, params)


class AgentTask(AgentOp):
    def __init__(self, launcher, blocking=None, aborter=None,
                 aborter_blocking=None):
        self.launcher = launcher
        self.blocking = blocking
        self.aborter = aborter
        if aborter_blocking is None:
            aborter_blocking = blocking
        self.aborter_blocking = aborter_blocking
        self.docstring = launcher.__doc__

    def encoded(self):
        """Dict of static info for API self-description."""
        return {
            'blocking': self.blocking,
            'abortable': (self.aborter is not None),
            'docstring': self.docstring,
            'op_type': 'task',
        }


class AgentProcess(AgentOp):
    def __init__(self, launcher, stopper, blocking=None, stopper_blocking=None):
        self.launcher = launcher
        self.stopper = stopper
        self.blocking = blocking
        if stopper_blocking is None:
            stopper_blocking = blocking
        self.stopper_blocking = stopper_blocking
        self.docstring = launcher.__doc__

    def encoded(self):
        """Dict of static info for API self-description."""
        return {
            'blocking': self.blocking,
            'docstring': self.docstring,
            'op_type': 'process',
        }


#: These are the valid values for session.status.  Use like this:
#:
#: - None: uninitialized.
#: - ``starting``: the Operation code has been launched and is
#:   performing basic quick checks in anticipation of moving to the
#:   (longer term) "running" state.
#: - ``running``: the Operation code has performed basic quick checks
#:   and has started to do the requested thing.
#: - ``stopping``: the Operation code has acknowledged receipt of a
#:   "stop" or "abort" request.
#: - ``done``: the Operation has exited, either succesfully or not.
#:
SESSION_STATUS_CODES = [None, 'starting', 'running', 'stopping', 'done']


class OpSession:
    """When a caller requests that an Operation (Process or Task) is
    started, an OpSession object is created and is associated with
    that run of the Operation.  The Operation codes are given access
    to the OpSession object, and may update the status and post
    messages to the message buffer.  This is the preferred means for
    communicating Operation status to the caller.

    In the OCSAgent model, Operations may run in the main, "reactor"
    thread or in a worker "pool" thread.  Services provided by
    OpSession must support both these contexts (see, for example,
    add_message).

    Control Clients are given a copy of the latest session information
    in each response from the Operation API.  The format of that
    information is described in ``.encoded()``.

    The message buffer is purged periodically.

    """

    def __init__(self, session_id, op_name, status='starting',
                 app=None, purge_policy=None):
        # Note that some data members are used internally, while others are
        # communicated over WAMP to Agent control clients.

        self.messages = []  # entries are time-ordered (timestamp, text).
        self.data = {}      # Operation-specific data structures.
        self.degraded = False
        self.session_id = session_id
        self.op_name = op_name
        self.start_time = time.time()
        self.end_time = None
        self.app = app
        self.success = None
        self.status = None

        # This has to be the last call since it depends on init...
        self.set_status(status, timestamp=self.start_time)

        # Set up the log message purge.
        self.purge_policy = {
            'min_age_s': 3600,     # Time in seconds after which
                                   # messages can be discarded.
            'min_messages': 5,     # Number of messages to keep,
                                   # even if they have expired.
            'max_messages': 10000,  # Max number of messages to keep,
                                   # even if they have not expired.
        }
        if purge_policy is not None:
            self.purge_policy.update(purge_policy)
        self.purge_log()

    def purge_log(self):
        cutoff = time.time() - self.purge_policy['min_age_s']
        while ((len(self.messages) > self.purge_policy['max_messages'])
               or ((len(self.messages) > self.purge_policy['min_messages'])
               and self.messages[0][0] < cutoff)):
            self.messages.pop(0)
        # Set this purger to be called again in the future, at some
        # cadence based on the minimum message age.
        next_purge_time = max(self.purge_policy['min_age_s'] / 5, 600)
        self.purger = task.deferLater(reactor, next_purge_time, self.purge_log)

    def encoded(self):
        """Encode the session data in a dict.  This is the data structure that
        is returned to Control Clients using the Operation API, as the
        "session" information.  Note the returned object is a dict
        with entries described below.

        Returns
        -------
        session_id : int
          A unique identifier for a single session (a single "run" of
          the Operation).  When an Operation is initiated, a new
          session object is created and can be distinguished from
          other times the Operation has been run using this id.
        op_name : str
          The OCS Operation name.
        op_code : int
          The OpCode, which combines information from status, success,
          and degraded; see :class:`ocs.base.OpCode`.
        status : str
          The Operation run status (e.g. 'starting', 'done', ...).
          See :data:`ocs.ocs_agent.SESSION_STATUS_CODES`.
        degraded: bool
          A boolean flag (defaults to False) that an operation may set
          to indicate that it is not achieving its primary function
          (e.g. if it cannot establish connection to hardware).
        success : bool or None
          If the Operation Session has completed (`status == 'done'`),
          this indicates that the Operation was deemed successful.
          Prior to the completion of the operation, the value is None.
          The value could be False if the Operation reported failure,
          or if it crashed and failure was marked by the encapsulating
          OCS code.
        start_time : float
          The time the Operation Session started, as a unix timestamp.
        end_time : float or None
          The time the Operation Session ended, as a unix timestamp.
          While the Session is still on-going, this is None.
        data : dict
          This is an area for the Operation code to store custom
          information for Control Clients to consume.  See notes
          below.  This structure will be tested for strict JSON
          compliance, and certain translations performed (such as
          converting NaN to None/null).
        messages : list
          A buffer of messages posted by the Operation.  Each element
          of the list is a tuple, (timestamp, message) where timestamp
          is a unix timestamp and message is a string.

        Notes
        -----
        The ``data`` field may be used by Agent code to provide data
        that might be of interest to a user (whether human or
        automated), such as the most recent readings from a device,
        structured information about the configuration and progress of
        the Operation, or diagnostics.

        Please see developer documentation (:ref:`session_data`) for
        advice on structuring your Agent session data.

        """
        def json_safe(data, check_ok=False):
            """Convert data so it can be serialized and decoded on
            the other end.  This includes:

            - Converting numpy arrays and scalars to generic lists and
              Python basic types.

            - Converting NaN to None (although crossbar handles
              NaN/inf, web browsers may fail to deserialize the
              invalid JSON this requires).

            In the case of inf/-inf, a ValueError is raised.

            """
            if check_ok:
                output = json_safe(data)
                json.dumps(output, allow_nan=False)
                return output
            if isinstance(data, dict):
                return {k: json_safe(v) for k, v in data.items()}
            if isinstance(data, (list, tuple)):
                return [json_safe(x) for x in data]
            if hasattr(data, 'dtype'):
                # numpy arrays and scalars.
                return json_safe(data.tolist())
            if isinstance(data, (str, int, bool)):
                return data
            if isinstance(data, float):
                if math.isnan(data):
                    return None
                if not math.isfinite(data):
                    raise ValueError('Session.data cannot store inf/-inf; '
                                     'please convert to NaN.')
            # This could still be something weird but json.dumps will
            # probably reject it!
            return data

        return {'session_id': self.session_id,
                'op_name': self.op_name,
                'op_code': self.op_code.value,
                'status': self.status,
                'degraded': self.degraded,
                'success': self.success,
                'start_time': self.start_time,
                'end_time': self.end_time,
                'data': json_safe(self.data, True),
                'messages': self.messages}

    @property
    def op_code(self):
        """
        Returns the OpCode for the given session.  This is what will be
        published to the registry's ``operation_status`` feed.
        """
        if self.status is None:
            return OpCode.NONE
        elif self.status == 'starting':
            return OpCode.STARTING
        elif self.status == 'running':
            if self.degraded:
                return OpCode.DEGRADED
            else:
                return OpCode.RUNNING
        elif self.status == 'stopping':
            return OpCode.STOPPING
        elif self.success:
            return OpCode.SUCCEEDED
        else:
            return OpCode.FAILED

    def set_status(self, status, timestamp=None):
        """Update the OpSession status and possibly post a message about it.

        Args:
            status (string): New value for status (see below).
            timestamp (float): timestamp for the operation.

        The possible values for status are:

        'starting'
            This status object has just been created, and the
            Operation launch code has yet to run.

        'running'
            The Operation is running.

        'stopping'
            The Operation is running, but a stop or abort has been
            requested.

        'done'
            The Operation is has terminated.  (Failure / success must
            be determined separately.)

        The only valid transitions are forward in the sequence
        [starting, running, stopping, done]; i.e. it is forbidden for
        the status of an OpSession to move from stopping to running.

        If this function is called from a worker thread, it will be
        scheduled to run in the reactor, and will block until that is
        complete.

        """
        if timestamp is None:
            timestamp = time.time()
        if not in_reactor_context():
            return threads.blockingCallFromThread(reactor,
                                                  self.set_status, status,
                                                  timestamp=timestamp)
        # Sanity check the status value.
        from_index = SESSION_STATUS_CODES.index(self.status)  # current status valid?
        to_index = SESSION_STATUS_CODES.index(status)        # new status valid?
        assert (to_index >= from_index)  # Only forward moves in status are permitted.

        if to_index == from_index:
            return

        self.status = status
        if status == 'done':
            self.end_time = timestamp

        try:
            self.add_message('Status is now "%s".' % status, timestamp=timestamp)
        except (TransportLost, Disconnected):
            self.app.log.error('setting session status to "{s}" failed. '
                               + 'transport lost or disconnected', s=status)

    def add_message(self, message, timestamp=None):
        """Add a log message to the OpSession messages buffer.

        Args:
            message (string): Message to append.
            timestamp (float): timestamp to tag the message.  The
                default, which is None, will cause the timestamp to be
                computed here and should be used in most cases.

        """
        if timestamp is None:
            timestamp = time.time()
        if not in_reactor_context():
            return reactor.callFromThread(self.add_message, message,
                                          timestamp=timestamp)
        self.messages.append((timestamp, message))
        # Make the app log this message, too.  The op_name and
        # session_id are an important provenance prefix.
        self.app.log.info('%s:%i %s' % (self.op_name, self.session_id, message))


class ParamError(Exception):
    def __init__(self, msg):
        self.msg = msg


class ParamHandler:
    """Helper for Agent Operation codes to extract params.  Supports type
    checking, has casting, and will raise errors that are
    automatically added to the session log and propagated to the
    caller in a useful way.

    There are two ways to use this.  The first and recommended way is
    to use the @param decorator.  Example::

        from ocs import ocs_agent

        class MyAgent:
            ...

            @ocs_agent.param('voltage', type=float)
            @ocs_agent.param('delay_time', default=1., type=float)
            @ocs_agent.param('other_action', default=None, cast=str)
            def my_task(self, session, params):
                # (Type checking and default substitution have been done already)
                voltage = params['voltage']
                delay_time = params['delay_time']
                other_action = params['other_action']
                ...

    When you use the @param decorator, the OCS code can check the
    parameters immediately when they are received from the client, and
    immediately return an error message to the client's start request
    (without even calling the Op start function)::

        OCSReply: ERROR : Param 'delay'=two_seconds is not of required type (<class 'float'>)
           (no session -- op has never run)


    A second possibility is to instantiate a ParamHandler at the start
    of your Op start function, and use it to extract parameters.
    Example::

        from ocs import ocs_agent

        class MyAgent:
            ...
            def my_task(self, session, params):
                params = ocs_agent.ParamHandler(params)
                # Mandatory, and cannot be None.
                voltage = params.get('voltage', type=float)
                # Optional, defaults to 1.
                delay_time = params.get('delay_time', default=1., type=float)
                # Optional, interpret as string, but defaults to None.
                other_action = params.get('other_action', default=None, cast=str)
                ...

    In this case, errors will not be immediatley returned to the user,
    but the Operation will quickly fail, and the error message will
    show up in the message log::

        OCSReply: OK : Operation "my_task" is currently not running (FAILED).
          my_task[session=1]; status=done with ERROR 0.115665 s ago, took 0.000864 s
          messages (4 of 4):
            1629464204.780 Status is now "starting".
            1629464204.780 Status is now "running".
            1629464204.781 ERROR: Param 'delay'=two_seconds is not of required type (<class 'float'>)
            1629464204.781 Status is now "done".
          other keys in .session: op_code, data

    """

    def __init__(self, params):
        if params is None:
            params = {}
        self._params = params
        self._checked = set()

    def get(self, key, default=ParamError(''), check=None, cast=None, type=None,
            choices=None, treat_none_as_missing=True):
        """Retrieve a value from the wrapped params dict, with optional type
        checking, casting, and validity checks.  If a parameter is
        found to be missing, or its value not valid, then a ParamError
        is raised.

        In Agent Op implementations, the ParamError will be caught by
        the API wrapper and automatically propagated to the caller.
        The Operation session will be marked as "done", with
        success=False.

        This works best if the implementation validates all parameters
        *before* beginning any Operation activities!

        Args
        ----
        key : str
          The name of the parameter to extract.
        default : any
          The value to use if the value is not set.  If this isn't
          explicitly set, then a missing key causes an error to be
          raised (see also the treat_none_as_missing arg).
        check : callable
          A function that will validate the argument; if the function
          returns False then a ParamError will be raised.
        cast : callable
          A function to run on the value to convert it.  For example
          ``cast=str.lower`` would help convert user argument
          "Voltage" to value "voltage".
        type : type
          Class to which the result will be compared, unless it is
          ``None``.  Note that if you pass ``type=float``, ``int``
          values will automatically be cast to ``float`` and accepted
          as valid.
        choices : list
          Acceptable values for the parameter.  This is checked after
          casting.
        treat_none_as_missing : bool
          Determines whether a value of ``None`` for a parameter
          should be treated in the same way as if the parameter were
          not set at all.  See notes.

        Returns
        -------
        The fully processed value.

        Notes
        -----

        The default behavior is to treat ``{'param': None}`` as the
        same as ``{}``; i.e. passing ``None`` as the value for a
        parameter is the same as leaving the parameter unset.  In both
        of these cases, unless a ``default=...`` is specified, a
        ``ParamError`` will be raised.  Note this doesn't preclude you
        from setting ``default=None``, which would effectively convert
        ``{}`` to ``{'param': None}``.  If you really need to block
        ``{}`` while allowing ``{'param': None}`` to be taken at face
        value, then set ``treat_none_as_missing=False``.

        The cast function, if specified, is applied before the type,
        choices, and check arguments are processed.  If the value (or
        the substituted default value) is ``None``, then any specified
        cast and checks will not be performed.

        """
        self._checked.add(key)
        value = self._params.get(key, None)
        is_unset = value is None and \
            (treat_none_as_missing or key not in self._params)
        if is_unset:
            if isinstance(default, ParamError):
                raise ParamError(f"Param '{key}' is required and must not be None")
            value = default
        if value is not None:
            if cast is not None:
                try:
                    value = cast(value)
                except BaseException:
                    raise ParamError(f"Param '{key}'={value} could not be cast to {cast}.")
            if type is not None:
                # Free cast from int to float.
                if type is float and isinstance(value, int):
                    value = float(value)
                # Fix type after json conversion
                if type is tuple and isinstance(value, list) and cast in [tuple, None]:
                    value = tuple(value)
                if not isinstance(value, type):
                    raise ParamError(f"Param '{key}'={value} is not of required type ({type})")
            if choices is not None:
                if value not in choices:
                    raise ParamError(f"Param '{key}'={value} is not in allowed set ({choices})")
            if check is not None:
                if not check(value):
                    raise ParamError(f"Param '{key}' failed validity check (see docs?).")
        return value

    def batch(self, instructions, check_for_strays=True):
        """
        Supports the @params decorator ... see code.
        """
        params = {}
        for key, kw in instructions:
            if key == '_':
                pass
            elif key == '_no_check_strays':
                check_for_strays = False
            else:
                params[key] = self.get(key, **kw)
        if check_for_strays:
            self.check_for_strays()
        return params

    def check_for_strays(self, ignore=[]):
        """Raise a ParamError if there were arguments passed in that have not
        yet been extracted with .get().  Keys passed in ignore (list)
        will be ignored regardless.

        """
        weird_args = [k for k in self._params.keys()
                      if k not in self._checked and k not in ignore]
        if len(weird_args):
            raise ParamError(f"params included unexpected values: {weird_args}")


def param(key, **kwargs):
    """Decorator for Agent operation functions to assist with checking
    params prior to actually trying to execute the code.  Example::

      class MyAgent:
        ...
        @param('voltage', type=float)
        @param('delay', default=0., type=float)
        @inlineCallbacks
        def set_voltage(self, session, params):
          ...

    Note the ``@param`` decorators should be all together, and
    outermost (listed first).  This is because the current
    implementation caches data in the decorated function (or
    generator) directly, and additional decorators will conceal that.

    See :class:`ocs.ocs_agent.ParamHandler` for more details.  Note the
    signature for @param is the same as for :func:`ParamHandler.get`.

    """
    # Validate the kwargs by passing them to "get" with trivial data.
    if 'default' in kwargs:
        ParamHandler({}).get(key, **kwargs)
    else:
        ParamHandler({}).get(key, default=None, **kwargs)
    # Start a cache and append these args to it...

    def deco(func):
        if not hasattr(func, '_ocs_prescreen'):
            setattr(func, '_ocs_prescreen', [])
        func._ocs_prescreen.append((key, kwargs))
        return func
    return deco



"""
Register our agents in ocs central.  In order for this script to
be imported by site_config.scan_for_agents(), it must be in the python
path and called something like ocs_plugin_*.
"""

import ocs
import os
root = os.path.abspath(os.path.split(__file__)[0])

for n, f in [
        ('RegistryAgent', 'registry/agent.py'),
        ('AggregatorAgent', 'aggregator/agent.py'),
        ('HostManager', 'host_manager/agent.py'),
        ('FakeDataAgent', 'fake_data/agent.py'),
        ('InfluxDBAgent', 'influxdb_publisher/agent.py'),
        ('BarebonesAgent', 'barebones/agent.py'),
]:
    ocs.site_config.register_agent_class(n, os.path.join(root, f))



import time
import txaio

from os import environ

from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock


class BarebonesAgent:
    """Barebone Agent demonstrating writing an Agent from scratch.

    This Agent is meant to be an example for Agent development, and provides a
    clean starting point when developing a new Agent.

    Parameters:
        agent (OCSAgent): OCSAgent object from :func:`ocs.ocs_agent.init_site_agent`.

    Attributes:
        agent (OCSAgent): OCSAgent object from :func:`ocs.ocs_agent.init_site_agent`.
        log (txaio.tx.Logger): Logger object used to log events within the
            Agent.
        lock (TimeoutLock): TimeoutLock object used to prevent simultaneous
            commands being sent to hardware.
        _count (bool): Internal tracking of whether the Agent should be
            counting or not. This is used to exit the Process loop by changing
            it to False via the count.stop() command. Your Agent won't use this
            exact attribute, but might have a similar one.

    """

    def __init__(self, agent):
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock(default_timeout=5)
        self._count = False

        # Register OCS feed
        agg_params = {
            'frame_length': 10 * 60  # [sec]
        }
        self.agent.register_feed('feed_name',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1.)

    def count(self, session, params):
        """count(test_mode=False)

        **Process** - Count up from 0.

        The count will restart if the process is stopped and restarted.

        Notes:
            The most recent value is stored in the session data object in the
            format::

                >>> response.session['data']
                {"value": 0,
                 "timestamp":1600448753.9288929}

        """
        with self.lock.acquire_timeout(timeout=0, job='count') as acquired:
            if not acquired:
                print("Lock could not be acquired because it "
                      + f"is held by {self.lock.job}")
                return False

            # Initialize last release time for lock
            last_release = time.time()

            # Initialize the counter
            self._count = True
            counter = 0

            self.log.info("Starting the count!")

            # Main process loop
            while self._count:
                # About every second, release and acquire the lock
                if time.time() - last_release > 1.:
                    last_release = time.time()
                    if not self.lock.release_and_acquire(timeout=10):
                        print(f"Could not re-acquire lock now held by {self.lock.job}.")
                        return False

                # Perform the process actions
                counter += 1
                self.log.debug(f"{counter}! Ah! Ah! Ah!")
                now = time.time()
                session.data = {"value": counter,
                                "timestamp": now}

                # Format message for publishing to Feed
                message = {'block_name': 'count',
                           'timestamp': now,
                           'data': {'value': counter}}
                self.agent.publish_to_feed('feed_name', message)
                time.sleep(1)

        self.agent.feeds['feed_name'].flush_buffer()

        return True, 'Acquisition exited cleanly.'

    def _stop_count(self, session, params):
        """Stop monitoring the turbo output."""
        if self._count:
            self._count = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'count is not currently running'

    @ocs_agent.param('text', default='hello world', type=str)
    def print(self, session, params):
        """print(text='hello world')

        **Task** - Print some text passed to a Task.

        Args:
            text (str): Text to print out. Defaults to 'hello world'.

        Notes:
            The session data will be updated with the text::

                >>> response.session['data']
                {'text': 'hello world',
                 'last_updated': 1660249321.8729222}

        """
        with self.lock.acquire_timeout(timeout=3.0, job='print') as acquired:
            if not acquired:
                self.log.warn("Lock could not be acquired because it "
                              + f"is held by {self.lock.job}")
                return False

            # Log the text provided to the Agent logs
            self.log.info(f"{params['text']}")

            # Store the text provided in session.data
            session.data = {'text': params['text'],
                            'last_updated': time.time()}

        # bool, 'descriptive text message'
        # True if task succeeds, False if not
        return True, 'Printed text to logs'


def add_agent_args(parser_in=None):
    if parser_in is None:
        from argparse import ArgumentParser as A
        parser_in = A()
    pgroup = parser_in.add_argument_group('Agent Options')
    pgroup.add_argument('--mode', type=str, default='count',
                        choices=['idle', 'count'],
                        help="Starting action for the Agent.")

    return parser_in


def main(args=None):
    # For logging
    txaio.use_twisted()
    txaio.make_logger()

    # Start logging
    txaio.start_logging(level=environ.get("LOGLEVEL", "info"))

    parser = add_agent_args()
    args = site_config.parse_args(agent_class='BarebonesAgent',
                                  parser=parser,
                                  args=args)

    startup = False
    if args.mode == 'count':
        startup = True

    agent, runner = ocs_agent.init_site_agent(args)

    barebone = BarebonesAgent(agent)
    agent.register_process(
        'count',
        barebone.count,
        barebone._stop_count,
        startup=startup)
    agent.register_task('print', barebone.print)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import time
import queue
import argparse
import txaio

from twisted.internet import reactor

from os import environ
from ocs import ocs_agent, site_config
from ocs.base import OpCode

from ocs.agents.aggregator.drivers import Aggregator

# For logging
txaio.use_twisted()
LOG = txaio.make_logger()


class AggregatorAgent:
    """
    This class provide a WAMP wrapper for the data aggregator. The run function
    and the data handler **are** thread-safe, as long as multiple run functions
    are not started at the same time, which should be prevented through OCSAgent.

    Args:
        agent (OCSAgent):
            OCS Agent object
        args (namespace):
            args from the function's argparser.

    Attributes:
        time_per_file (int):
            Time (sec) before files should be rotated.
        data_dir (path):
            Path to the base directory where data should be written.
        aggregate (bool):
           Specifies if the agent is currently aggregating data.
        incoming_data (queue.Queue):
            Thread-safe queue where incoming (data, feed) pairs are stored before
            being passed to the Aggregator.
        loop_time (float):
            Time between iterations of the run loop.
    """

    def __init__(self, agent, args):
        self.agent: ocs_agent.OCSAgent = agent
        self.log = agent.log

        self.time_per_file = int(args.time_per_file)
        self.data_dir = args.data_dir

        self.aggregate = False
        self.incoming_data = queue.Queue()
        self.loop_time = 1

        # SUBSCRIBES TO ALL FEEDS!!!!
        # If this ends up being too much data, we can add a tag '.record'
        # at the end of the address of recorded feeds, and filter by that.
        self.agent.subscribe_on_start(self._enqueue_incoming_data,
                                      f'{args.address_root}..feeds.',
                                      options={'match': 'wildcard'})

        record_on_start = (args.initial_state == 'record')
        self.agent.register_process('record',
                                    self.record, self._stop_record,
                                    startup=record_on_start)

    def _enqueue_incoming_data(self, _data):
        """
        Data handler for all feeds. This checks to see if the feeds should
        be recorded, and if they are it puts them into the incoming_data queue
        to be processed by the Aggregator during the next run iteration.
        """
        data, feed = _data

        if not feed['record'] or not self.aggregate:
            return

        self.incoming_data.put((data, feed))
        self.log.debug("Enqueued {d} from Feed {f}", d=data, f=feed)

    @ocs_agent.param('test_mode', default=False, type=bool)
    def record(self, session: ocs_agent.OpSession, params):
        """record(test_mode=False)

        **Process** - This process will create an Aggregator instance, which
        will collect and write provider data to disk as long as this process is
        running.

        Parameters:
            test_mode (bool, optional): Run the record Process loop only once.
                This is meant only for testing. Default is False.

        Notes:
            The most recent file and active providers will be returned in the
            session data::

                >>> response.session['data']
                {"current_file": "/data/16020/1602089117.g3",
                 "providers": {
                    "observatory.fake-data1.feeds.false_temperatures": {
                        "last_refresh": 1602089118.8225083,
                        "sessid": "1602088928.8294137",
                        "stale": false,
                        "last_block_received": "temps"},
                    "observatory.LSSIM.feeds.temperatures": {
                         "last_refresh": 1602089118.8223345,
                         "sessid": "1602088932.335811",
                         "stale": false,
                         "last_block_received": "temps"}}}

        """
        self.aggregate = True

        try:
            aggregator = Aggregator(
                self.incoming_data,
                self.time_per_file,
                self.data_dir,
                session=session
            )
        except PermissionError:
            self.log.error("Unable to intialize Aggregator due to permission "
                           "error, stopping twisted reactor")
            reactor.callFromThread(reactor.stop)
            return False, "Aggregation not started"

        while self.aggregate:
            time.sleep(self.loop_time)
            aggregator.run()

            if params['test_mode']:
                break

        aggregator.close()

        return True, "Aggregation has ended"

    def _stop_record(self, session, params):
        if OpCode(session.op_code) in [OpCode.STARTING, OpCode.RUNNING]:
            session.set_status('stopping')
            self.aggregate = False
            return True, "Stopping aggregation"
        elif OpCode(session.op_code) == OpCode.STOPPING:
            return True, "record process status is already 'stopping'"
        else:
            return False, "record process not currently running"


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--data-dir', required=True,
                        help="Base directory to store data. "
                             "Subdirectories will be made here.")
    pgroup.add_argument('--initial-state',
                        default='idle', choices=['idle', 'record'],
                        help="Initial state of argument parser. Can be either"
                             "idle or record")
    pgroup.add_argument('--time-per-file', default='3600',
                        help="Time per file in seconds. Defaults to 1 hr")

    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=environ.get("LOGLEVEL", "info"))

    parser = make_parser()
    args = site_config.parse_args(agent_class='AggregatorAgent',
                                  parser=parser,
                                  args=args)
    agent, runner = ocs_agent.init_site_agent(args)

    AggregatorAgent(agent, args)
    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



from enum import Enum


class ResponseCode(Enum):
    """Enumeration of response codes from the Operation API (start, stop,
    wait, ...).

    These response codes indicate only whether the API call was
    successful, in that the request was propagated all the way to the
    Agent's Operation code.  They are not used to represent success or
    failure of the Operation itself.

    """

    #: OK indicates the request was successful.
    OK = 0

    #: ERROR indicates that the request could not be propagated fully.
    #: This may occur, for example, if an invalid Operation name is
    #: passed, if a request is made that conflicts with an Operation's
    #: current state (e.g. .start() is called on an already-running
    #: Operation), if an API call is made to a Operation of an
    #: incompatible type (e.g. .stop() on a Task), or due to API
    #: syntax error (e.g. misspelled keyword argument to .wait()).
    ERROR = -1

    #: TIMEOUT is returned in the case that a Client issued a blocking
    #: call with timeout (.wait()), and the timeout expired before the
    #: Operation completed.
    TIMEOUT = 1


OK = ResponseCode.OK.value
ERROR = ResponseCode.ERROR.value
TIMEOUT = ResponseCode.TIMEOUT.value


class OpCode(Enum):
    """Enumeration of OpSession "op_code" values.

    The op_code corresponds to the session.status, with the following
    extensions:

    - If the session.status == "done" then the op_code will be
      assigned a value of either SUCCEEDED or FAILED based on
      session.success.
    - If the session.status == "running", and session.degraded is
      True, then the op_code will be DEGRADED rather than RUNNING.

    """

    #: NONE is used to represent an uninitialized OpSession, and does
    #: not correspond to some attempt to run the Operation.
    NONE = 1

    #: STARTING indicates that start() has been successfully called,
    #: but the Operation has not yet marked itself as successfully
    #: launched.  If this state is reached, then at the very least the
    #: start request was not rejected because it was already running.
    STARTING = 2

    #: RUNNING indicates that the Operation has performed its basic
    #: initialization and parameter checking and is performing its
    #: task.  Operation codes need to explicitly mark themselves as
    #: running by calling session.set_state('running').
    RUNNING = 3

    #: STOPPING indicates that the Agent has received a stop or abort
    #: request for this Operation and will try to wrap things up ASAP.
    STOPPING = 4

    #: SUCCEEDED indicates that the Operation has terminated and has
    #: indicated the Operation was successful.  This includes the case
    #: of a Process that was asked to stop and has shut down cleanly.
    SUCCEEDED = 5

    #: FAILED indicates that the Operation has terminated with some
    #: kind of error.
    FAILED = 6

    #: EXPIRED may used to mark session information as invalid in cases
    #: where the state cannot be determined.
    EXPIRED = 7

    #: DEGRADED indicates that an operation meets the requirements for
    #: state RUNNING, but is self-reporting as being in a problematic
    #: state where it is unable to perform its primary functions (for
    #: example, if a Process to operate some hardware is trying to
    #: re-establish connection to that hardware).
    DEGRADED = 8



from .base import OK, ERROR, TIMEOUT, ResponseCode, OpCode  # noqa: F401

from . import site_config  # noqa: F401

# Define the variable '__version__':
# This has the closest behavior to versioneer that I could find
# https://github.com/maresb/hatch-vcs-footgun-example
try:
    # If setuptools_scm is installed (e.g. in a development environment with
    # an editable install), then use it to determine the version dynamically.
    from setuptools_scm import get_version

    # This will fail with LookupError if the package is not installed in
    # editable mode or if Git is not installed.
    __version__ = get_version(root="..", relative_to=__file__, version_scheme="no-guess-dev")
except (ImportError, LookupError):
    # As a fallback, use the version that is hard-coded in the file.
    try:
        from ocs._version import __version__  # noqa: F401
    except ModuleNotFoundError:
        # The user is probably trying to run this without having installed
        # the package, so complain.
        raise RuntimeError(
            "ocs is not correctly installed. "
            "Please install it with pip."
        )



import argparse
import importlib
import os
import setproctitle
import sys
import warnings

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

from ocs import site_config

DESCRIPTION = """
This script provides a quick way to start an OCS Agent. You will need to set
``OCS_CONFIG_DIR`` environment variable to the directory containing
default.yaml, or else use ``--site-*`` options to specify your configuration.

To start an Agent, run::

  ocs-agent-cli --instance-id INSTANCE_ID

``ocs-agent-cli`` will also inspect environment variables for commonly
passed arguments to facilitate configuration within Docker containers.
Those environment variables, if defined and non-trivial, will be
passed to the agent script unless they are overridden explicitly on
the ``ocs-agent-cli`` command line.  The environment variables and
arguments are:

  - ``INSTANCE_ID``, will be the default value for ``--instance-id``
  - ``SITE_HOST``, for ``--site-host``
  - ``SITE_HUB``, for ``--site-hub``
  - ``SITE_HTTP``, for ``--site-http``
  - ``CROSSBAR_TIMEOUT``, for ``--crossbar-timeout``


``ocs-agent-cli`` relies on the Agent being run belonging to an OCS Plugin. If
the Agent is not an OCS Plugin it can be run directly using both the
``--agent`` and ``--entrypoint`` flags. For example::

  ocs-agent-cli --agent my_agent.py --entrypoint main --instance-id my-agent-1

"""


def _get_parser():
    parser = argparse.ArgumentParser(
        description=DESCRIPTION,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    # Passed through to Agent
    parser.add_argument('--instance-id', default=None, help="Agent unique instance-id. E.g. 'aggregator' or 'fakedata-1'.")
    parser.add_argument('--site-hub', default=None, help="Site hub address.")
    parser.add_argument('--site-http', default=None, help="Site HTTP address.")
    parser.add_argument('--site-host', default=None, help="Declare the host the instance is configured in.")
    # Default set in site_config.py within add_arguments()
    parser.add_argument('--crossbar-timeout', type=int, help="Length of time in seconds "
                        "that the Agent will try to reconnect to the crossbar server before "
                        "shutting down. Disable the timeout by setting to 0.")

    # Not passed through to Agent
    parser.add_argument('--agent', default=None, help="Path to non-plugin OCS Agent.")
    parser.add_argument('--entrypoint', default=None, help="Agent module entrypoint function.")

    return parser


def build_agent_list():
    """Builds a list of all Agents available across all ocs plugins installed
    on the system.

    Note:
        Currently if two plugins provide the same Agent the one loaded first is
        used. This should be improved somehow if we expect overlapping Agents
        to be provided by plugins.

    Examples:
        An example agent list:

        >>> build_agent_list()
        {'RegistryAgent': {'module': 'ocs.agents.registry.agent', 'entry_point': 'main'},
        'AggregatorAgent': {'module': 'ocs.agents.aggregator.agent', 'entry_point': 'main'},
        'HostManager': {'module': 'ocs.agents.host_manager.agent', 'entry_point': 'main'},
        'FakeDataAgent': {'module': 'ocs.agents.fake_data.agent', 'entry_point': 'main'},
        'InfluxDBAgent': {'module': 'ocs.agents.influxdb_publisher.agent', 'entry_point': 'main'},
        'BarebonesAgent': {'module': 'ocs.agents.barebones.agent', 'entry_point': 'main'}}

    Returns:
        dict: Dictionary of available agents, with agent names as the keys, and
        dicts containing the module and entry_point as values.

    """
    discovered_plugins = entry_points(group='ocs.plugins')
    print("Installed OCS Plugins:", [x.name for x in discovered_plugins])
    agents = {}
    for name in discovered_plugins.names:
        (plugin, ) = discovered_plugins.select(name=name)
        try:
            loaded = plugin.load()
        except Exception as e:
            print(f"Could not load plugin: {name}")
            print("  Error:", e)
            continue

        # Remove any duplicate agent classes from newly loaded plugin
        current_agents = set(agents)
        conflicts = current_agents.intersection(set(loaded.agents))
        for con in conflicts:
            warnings.warn(
                f'Found duplicate agent-class {con} provided by {name}. '
                + f'Using {agents[con]}.')
            del loaded.agents[con]

        agents.update(loaded.agents)

    return agents


def main(args=None):
    # Grab commandline arguments
    if args is None:
        args = sys.argv[1:]

    parser = _get_parser()
    # pre_args relevant to ocs-agent-cli, post_args passed to Agent entrypoint
    pre_args, post_args = parser.parse_known_args(args)

    # Required arguments
    if pre_args.instance_id is not None:
        # Re-add instance-id to post_args
        post_args.extend(["--instance-id", pre_args.instance_id])
    else:
        # Grab instance-id from ENV for running in Docker
        id_env = os.environ.get("INSTANCE_ID", None)

        # Inject ENV based instance-id only if not passed on cli
        post_args.extend(["--instance-id", id_env])

        # instance-id is always required
        if id_env is None:
            print("--instance-id (or $INSTANCE_ID) not provided. Exiting.")
            sys.exit(1)

    # Optional arguments
    # Add additional optional arguments by appending to optional_env
    # Format is {"arg name within argparse": "ENVIRONMENT VARIABLE NAME"}
    # E.g. --my-new-arg should be {"my_new_arg": "MY_NEW_ARG"}
    optional_env = {"site_hub": "SITE_HUB",
                    "site_http": "SITE_HTTP",
                    "site_host": "SITE_HOST",
                    "crossbar_timeout": "CROSSBAR_TIMEOUT",
                    }

    for _name, _var in optional_env.items():
        # Args passed on cli take priority
        _arg = vars(pre_args)[_name]
        _flag = f"--{_name}".replace('_', '-')
        if _arg is not None:
            post_args.extend([_flag, str(_arg)])
            continue

        # Get from ENV if not passed w/flag
        set_var = os.environ.get(_var, None)
        if set_var is not None:
            post_args.extend([_flag, set_var])

    print('Args:', post_args)  # mostly to debug, but also useful
    # Handle running agent directly (outside of plugin package)
    if pre_args.agent:
        agent_file = pre_args.agent

        # Entrypoint required if running outside of plugin system
        if pre_args.entrypoint is None:
            print("--entrypoint not provided. Exiting")
            sys.exit(1)

        entrypoint = pre_args.entrypoint

        # Insert agent into path for import
        script_path = os.path.dirname(agent_file)
        sys.path.insert(0, script_path)
        mod = importlib.import_module(os.path.basename(agent_file)[:-3])
    else:
        # Note this is only used to lookup the agent-class, post_args are
        # passed to the agent's entrypoint below when calling start(), which
        # includes injected ENV based arguments.
        args = site_config.parse_args(agent_class='*host*',
                                      parser=None, args=post_args)

        # Determine agent-class from instance-id
        (site_, host_, instance) = site_config.get_config(args)
        agent_class = instance.data['agent-class']

        # Import agent's entrypoint and execute
        agents = build_agent_list()
        agent_info = agents[agent_class]
        _module = agent_info["module"]
        entrypoint = agent_info.get("entry_point", "main")

        mod = importlib.import_module(_module)

        title = f'ocs-agent:{instance.data["instance-id"]}'
        print(f'Renaming this process to: "{title}"')
        setproctitle.setproctitle(title)

    start = getattr(mod, entrypoint)  # This is the start function.
    start(args=post_args)


if __name__ == '__main__':
    # This __main__ checker allows ocs-agent-cli to be invoked through
    # python -m ocs.agent_cli; this is helpful with, e.g., conda when
    # one is trying to use a particular interpreter and its packages.
    main()



import os
import datetime


def _find_all_g3_files(target):
    """Build list of .g3 files.

    Parameters
    ----------
    target : str
        File or directory to scan.

    Returns
    -------
    list
        List of full paths to .g3 files.

    """
    _file_list = []
    if os.path.isfile(target):
        _file_list.append(target)
    elif os.path.isdir(target):
        a = os.walk(target)
        for root, _, _file in a:
            for g3 in _file:
                if g3[-2:] == "g3":
                    _file_list.append(os.path.join(root, g3))

    return _file_list


def _remove_non_matching_files(filelist, verbose=None):
    """Make new list, removing files that don't match the known stringtime format.

    If the script was partially run on a directory, newly renamed files will
    cause errors at the strptime step, so remove them.

    Parameters
    ----------
    filelist : list
        List of files, probably made by _find_all_g3_files()
    verbose : int
        Verbosity level.

    Returns
    -------
    list
        List of full paths to .g3 files.

    """
    new_list = []
    for f in filelist:
        try:
            basename = os.path.basename(f)
            datetime.datetime.strptime(basename, "%Y-%m-%d-%H-%M-%S.g3")
            new_list.append(f)
        except ValueError:
            if verbose is not None:
                print(f"{f} does not match datestring format, removing from list.")

    return new_list


def build_filelist(target, verbose=None):
    """Build .g3 filelist to process.

    This is done in two steps, finding all *.g3 files, then making sure they
    match the old naming convention.

    Parameters
    ----------
    target : str
        File or directory to scan.
    verbose : int
        Verbosity level.

    Returns
    -------
    list
        List of full paths to .g3 files.

    """

    _filelist = _find_all_g3_files(target)
    filelist = _remove_non_matching_files(_filelist, verbose)

    return filelist


def _generate_ctime_filename(path):
    """Determine the ctime corresponding to a datestring formatted filename.

    Note: Files must be named with a UTC timestamp for this to work properly.
    However, the code works on any timezone machine. If files are not named based
    on UTC timestamps, then this will result in a shifted timestamp after
    conversion.

    Paramters
    ---------
    path : str
        Full path to old datestring style .g3 file.

    Returns
    -------
    str
        File basename in new ctime based format.

    """
    basename = os.path.basename(path)
    dt = datetime.datetime.strptime(basename, "%Y-%m-%d-%H-%M-%S.g3")
    ctime = int(dt.replace(tzinfo=datetime.timezone.utc).timestamp())
    new_filename = f"{ctime}.g3"

    return new_filename


def _rename_file(path, verbose=None):
    """Rename a single file from "%Y-%m-%d-%H-%M-%S.g3" to "ctime.g3".

    Paramters
    ---------
    path : str
        Filename for .g3 file.
    verbose : int
        Verbosity level.

    """
    dirname = os.path.dirname(path)
    new_filename = _generate_ctime_filename(path)
    new_path = os.path.join(dirname, new_filename)
    if verbose is not None:
        print(f"Renaming {path} to {new_path}")

    # Don't overwrite any existing file.
    if os.path.isfile(new_path):
        raise OSError("File at dst, {new_path}, already exists!")

    os.rename(path, new_path)


def rename_files(path, verbose=None):
    """Rename all .g3 file in path to new ctime based format.

    Paramters
    ---------
    path : str
        Filename for .g3 file.
    verbose : int
        Verbosity level.

    """
    filelist = build_filelist(path, verbose)

    for f in filelist:
        _rename_file(f, verbose)


def main(target, verbose=None):
    rename_files(target, verbose)



import collections
import time

import ocs
from ocs import site_config


def _get_op(op_type, name, encoded, client):
    """Factory for generating matched operations. This will make sure
    op.start's docstring is the docstring of the operation.

    Parameters:
        op_type (str): Operation type, either 'task' or 'process'.
        name (str): Operation name
        encoded (dict): Encoded :class:`ocs.ocs_agent.AgentTask` or
            :class:`ocs.ocs_agent.AgentProcess` dictionary.
        client (ControlClient): Client object, which will be used to issue the
            requests for operation actions.

    """
    class MatchedOp:
        def start(self, **kwargs):
            return OCSReply(*client.request('start', name, params=kwargs))

        def wait(self, timeout=None):
            return OCSReply(*client.request('wait', name, timeout=timeout))

        def status(self):
            return OCSReply(*client.request('status', name))

    class MatchedTask(MatchedOp):
        def abort(self):
            return OCSReply(*client.request('abort', name))

        def __call__(self, **kw):
            """Runs self.start(**kw) and, if that succeeds, self.wait()."""
            result = self.start(**kw)
            if result[0] != ocs.OK:
                return result
            return self.wait()

    class MatchedProcess(MatchedOp):
        def stop(self):
            return OCSReply(*client.request('stop', name))

        def __call__(self):
            """Equivalent to self.status()."""
            return self.status()

    MatchedOp.start.__doc__ = encoded['docstring']

    if op_type == 'task':
        return MatchedTask()
    elif op_type == 'process':
        return MatchedProcess()
    else:
        raise ValueError("op_type must be either 'task' or 'process'")


def _opname_to_attr(name):
    for c in ['-', ' ']:
        name = name.replace(c, '_')
    return name


class OCSClient:
    """The simple OCS Client, facilitating task/process calls.

    OCSClient makes an Agent's tasks/processes available as class attributes,
    making it easy to setup a client instance and call the associated Agent's
    tasks and processes.

    Example:
        This example sets up an OCSClient object and calls a FakeDataAgent's
        Task (delay_task) and process (acq)::

            >>> client = OCSClient('fake-data-1')
            >>> client.delay_task(delay=5)
            >>> client.acq.start()

    Attributes:
        instance_id (str): instance-id for agent to run

    """

    def __init__(self, instance_id, **kwargs):
        """
        Args:
            instance_id (str): Instance id for agent to run
            args (list or args object, optional):
                Takes in the parser arguments for the client.
                If None, pass an empty list.
                If list, reads in list elements as arguments.
                Defaults to None.

        .. note::
            For additional ``**kwargs`` see site_config.get_control_client.

        """
        if kwargs.get('args') is None:
            kwargs['args'] = []

        self._client = site_config.get_control_client(instance_id, **kwargs)
        self.instance_id = instance_id
        self._api = self._client.get_api()

        for name, _, encoded in self._api['tasks']:
            setattr(self, _opname_to_attr(name),
                    _get_op('task', name, encoded, self._client))

        for name, _, encoded in self._api['processes']:
            setattr(self, _opname_to_attr(name),
                    _get_op('process', name, encoded, self._client))

    def __repr__(self):
        return f"OCSClient('{self.instance_id}')"


def _humanized_time(t):
    if abs(t) < 1.:
        return '%.6f s' % t
    if abs(t) < 120:
        return '%.1f s' % t
    if abs(t) < 120 * 60:
        return '%.1f mins' % (t / 60)
    if abs(t) < 48 * 3600:
        return '%.1f hrs' % (t / 3600)
    return '%.1f days' % (t / 86400)


class OCSReply(collections.namedtuple('_OCSReply',
                                      ['status', 'msg', 'session'])):
    def __repr__(self):
        try:
            ok_str = ocs.ResponseCode(self.status).name
        except ValueError:
            ok_str = '???'
        text = 'OCSReply: %s : %s\n' % (ok_str, self.msg)
        if self.session is None or len(self.session.keys()) == 0:
            return text + '  (no session -- op has never run)'

        # try/fail in here so we can make assumptions about key
        # presence and bail out to a full dump if anything is weird.
        try:
            handled = ['op_name', 'session_id', 'status', 'start_time',
                       'end_time', 'messages', 'success']

            s = self.session
            run_str = 'status={status}'.format(**s)
            if s['status'] in ['starting', 'running']:
                run_str += ' for %s' % _humanized_time(
                    time.time() - s['start_time'])
            elif s['status'] == 'done':
                if s['success']:
                    run_str += ' without error'
                else:
                    run_str += ' with ERROR'
                run_str += ' %s ago, took %s' % (
                    _humanized_time(time.time() - s['end_time']),
                    _humanized_time(s['end_time'] - s['start_time']))
            text += ('  {op_name}[session={session_id}]; '
                     '{run_str}\n'.format(run_str=run_str, **s))
            messages = s.get('messages', [])
            if len(messages):
                to_show = min(5, len(messages))
                text += ('  messages (%i of %i):\n' % (to_show, len(messages)))
                for m in messages:
                    text += '    %.3f %s\n' % (m[0], m[1])

            also = [k for k in s.keys() if k not in handled]
            if len(also):
                text += ('  other keys in .session: ' + ', '.join(also))

        except Exception as e:
            text += ('\n  [session decode failed with exception: %s\n'
                     '  Here is everything in .session:\n %s\n]') \
                % (e.args, self.session)
        return text



import threading
from contextlib import contextmanager
import time
from autobahn.twisted.util import sleep as dsleep
from twisted.internet.defer import inlineCallbacks


class TimeoutLock:
    def __init__(self, default_timeout=0):
        """
        Locking mechanism to be used by OCS Agents.

        Args:
            default_timeout (float, optional):
                Sets the default timeout value for acquire calls.
                Defaults to 0.
        """
        self.job = None
        self._active = threading.Lock()
        self._next = threading.Lock()
        self._default_timeout = default_timeout

    def acquire(self, timeout=None, job=None):
        """
        Acquires main lock.

        Args:
            timeout (float, optional):
                Sets the timeout for lock acquisition.

                If set to 0 the acquire calls will be non-blocking and will
                immediately return the result.

                If set to any value greater than 0, this will block until
                the lock is acquired or the timeout (in seconds) has been
                reached.

                If set to -1 this call will block indefinitely until the lock
                has been acquired.

                If not set (default), it will use the TimeoutLock's
                default_value (which itself defaults to 0).

            job (string, optional):
                Job name to be associated with current lock acquisition.
                The current job is stored in self.job so that it can be
                inspected by other threads.

        Returns:
            result (bool): Whether or not lock acquisition was successful.
        """
        if timeout is None:
            timeout = self._default_timeout
        if timeout is None or timeout == 0.:
            kw = {'blocking': False}
        else:
            kw = {'blocking': True, 'timeout': timeout}
        result = False
        if self._next.acquire(**kw):
            if self._active.acquire(**kw):
                self.job = job
                result = True
            self._next.release()
        return result

    def release(self):
        """
        Releases an acquired lock.
        """
        self.job = None
        self._active.release()

    def release_and_acquire(self, timeout=None):
        """
        Releases and immediately reacquires a lock. Because this uses a two-lock
        system, it is guaranteed that at least one  blocking acquire call will
        be able to take the active lock before this function is able to
        re-acquire it. However no other ordering is guaranteed.
        """
        job = self.job
        self.release()
        return self.acquire(timeout=timeout, job=job)

    @contextmanager
    def acquire_timeout(self, timeout=None, job='unnamed'):
        """
        Context manager to acquire and hold a lock.

        Args:
            timeout (float, optional):
                Sets the timeout for lock acquisition. See the ``acquire``
                method documentation for details.

            job (string, optional):
                Job name to be associated with current lock acquisition.

        Returns:
            result (bool): Whether or not lock acquisition was successful.

        The following example will attempt to acquire the lock with a timeout
        of three seconds::

            lock = TimeoutLock()

            with lock.acquire_timeout(timeout=3.0, job='acq') as acquired:
                if not acquired:
                    print(f"Lock could not be acquired because it is held by {lock.job}")
                    return False

                print("Lock acquired!")
        """
        result = self.acquire(timeout=timeout, job=job)
        if result:
            try:
                yield result
            finally:
                self.release()
        else:
            yield result


def in_reactor_context():
    """
    Determine whether the current threading context is the twisted
    main (reactor) thread, or a worker pool thread.  Returns True if
    it's the main thread.  Will raise RuntimeError if the thread name
    is confusing.
    """
    t = threading.currentThread()
    if 'PoolThread' in t.name:
        return False
    if 'MainThread' in t.name:
        return True
    raise RuntimeError('Could not determine threading context: '
                       'currentThread.name="%s"' % t.name)


class Pacemaker:
    """
    The Pacemaker is a class to help Agents maintain a regular sampling rate
    in their processes. The Pacemaker class will correct for the time spent
    in the body of the process loop in it's sleep function. Additionally, if
    run with the ``quantize`` options, the pacemaker will attempt to snap samples
    to a temporal grid (starting on the second) so that different agents can
    remain relatively synchronized.

    Args:
        sample_freq (float):
            The sampling frequency for the pacemaker to enforce. This can be a
            float, however in order to use the ``quantize`` option it must be
            a whole number.
        quantize (bool):
            If True, the pacemaker will snap to a grid starting on the second.
            For instance, if ``sample_freq`` is 4 and ``quantize`` is set to
            True, the pacemaker will make it so samples will land close to
            ``int(second) + (0, 0.25, 0.5, 0.75)``.

    Here is an example of how the Pacemaker can be used keep a 3 Hz quantized
    sample rate::

        pm = Pacemaker(3, quantize=True)
        take_data = True:
        while take_data:
            pm.sleep()
            print("Acquiring thermometry data...")
            time.sleep(np.random.uniform(0, .3))
    """

    def __init__(self, sample_freq, quantize=False):
        self.sample_freq = sample_freq
        self.sample_time = 1. / self.sample_freq
        self.next_sample = time.time()
        self.quantize = quantize

        if quantize and (sample_freq % 1 != 0):
            raise ValueError("Quantization only works for frequencies that are whole numbers.")

    def _set_next_sample(self):
        self.next_sample = time.time() + self.sample_time
        if self.quantize:
            # Snaps "next_sample" to grid defined by sample_freq
            self.next_sample = (self.next_sample + self.sample_time / 2) \
                // self.sample_time * self.sample_time

    def sleep(self):
        """
        Sleeps until the next calculated sampling time.
        """
        now = time.time()
        if now < self.next_sample:
            time.sleep(self.next_sample - now)
        self._set_next_sample()

    @inlineCallbacks
    def dsleep(self):
        """
        Sleeps in a non-blocking way by returning the deferred created by
        twisted's sleep method.
        """
        now = time.time()
        if now < self.next_sample:
            yield dsleep(self.next_sample - now)
        self._set_next_sample()



import ocs

import shutil
import socket
import os
import sys
import yaml
import argparse
import collections
import deprecation


class SiteConfig:
    def __init__(self):
        self.hosts = {}
        self.hub = None
        self.source_file = None

    @classmethod
    def from_dict(cls, data):
        """Args:
            data: The configuration dictionary.

        The configuration dictionary should have the following elements:

        ``hub`` (required)
            Describes what WAMP server and realm Agents and Clients
            should use.

        ``hosts`` (required)
            A dictionary of HostConfig descriptions.  The keys in this
            dictionary can be real host names on the network,
            pseudo-host names, or the special value "localhost".

        A HostConfig marked for "localhost" will match any host that
        does not have an exact match in the hosts dictionary.  This
        should normally be used only in single-host test systems or
        examples.

        Client programs will normally (i.e., by default) try to load
        the HostConfig associated with the system hostname (that which
        is returned by socket.gethostname()). But this can be
        overridden easily, for example by using the ``--site-host``
        command line argument.  It is thus quite reasonable to use the
        hosts dictionary to hold a set of useful configurations
        indexed by a user-specified string (a pseudo-host).

        """
        self = cls()
        hosts = data.get('hosts')
        if hosts:
            for k, v in hosts.items():
                assert (k not in self.hosts)  # duplicate host name in config file!
                self.hosts[k] = HostConfig.from_dict(v, parent=self, name=k)
        self.hub = HubConfig.from_dict(data['hub'], parent=self)
        return self

    @classmethod
    def from_yaml(cls, filename):
        filename = os.path.abspath(filename)
        with open(filename) as f:
            data = yaml.safe_load(f)
        self = cls.from_dict(data)
        self.source_file = filename
        return self


class HostConfig:
    def __init__(self, name=None):
        self.instances = []
        self.name = name
        self.agent_paths = []
        self.log_dir = None
        self.working_dir = os.getcwd()
        self.crossbar_timeout = None

    @classmethod
    def from_dict(cls, data, parent=None, name=None):
        """Args:
            data: The configuration dictionary.
            parent: the SiteConfig from which this data was extracted
                (this is stored as self.parent, but not used).

        The configuration dictionary should have the following elements:

        ``agent-instances`` (required)
            A list of AgentConfig descriptions.

        ``agent-paths`` (optional)
            A list of additional paths where OCS is permitted to
            search for Agent plugin modules.

        ``crossbar`` (optional)
            Settings to assist with starting / stopping / monitoring a
            crossbar server running on this host.  There is a single
            crossbar server for an OCS network and thus this entry
            should be defined for at most one of the hosts in the site
            config file.  Note that setting this to None (or null)
            will disable host crossbar control, while setting it to an
            empty dictionary, {}, will enable local host crossbar
            control with default settings.

        ``log-dir`` (optional)
            Path at which to write log files.  Relative paths will be
            interpreted relative to the "working directory"; see
            --working-dir command line option.
        """
        self = cls(name=name)
        self.parent = parent
        self.data = data
        self.instances = data['agent-instances']
        self.agent_paths = data.get('agent-paths', [])
        self.crossbar = CrossbarConfig.from_dict(data.get('crossbar'))
        self.log_dir = data.get('log-dir', None)
        self.crossbar_timeout = data.get('crossbar_timeout', 10)
        return self


class CrossbarConfig:
    @classmethod
    def from_dict(cls, data, parent=None):
        """Args:
            data: The configuration dictionary, or None.
            parent: the HostConfig from which this data was extracted
                (this is stored as self.parent, but not used).

        The configuration dictionary should have the following elements:

        ``config-dir`` (optional): Location of crossbar config.json;
            this gets passed to ``--cbdir``, if specified..

        ``bin`` (optional): The path to the crossbar executable.
            This defaults to shutil.which('crossbar').

        If data is None, returns None.  Otherwise returns a
        CrossbarConfig object.
        """
        if data is None:
            return None
        self = cls()
        self.parent = parent
        self.binary = data.get('bin', shutil.which('crossbar'))
        self.cbdir = data.get('config-dir')
        if self.cbdir is None:
            self.cbdir_args = []
        else:
            self.cbdir_args = ['--cbdir', self.cbdir]
        return self

    def get_cmd(self, cmd):
        if self.binary is None:
            raise RuntimeError("Crossbar binary could not be found in PATH; "
                               "specify the binary in site_config?")
        if not os.path.exists(self.binary):
            raise RuntimeError("The crossbar binary specified in site_config "
                               "does not seem to exist: %s" % self.binary)
        return [self.binary, cmd] + self.cbdir_args

    def summary(self):
        return summarize_dict({
            'bin': self.binary,
            'config-dir': self.cbdir,
        })


class HubConfig:
    @classmethod
    def from_dict(cls, data, parent=None):
        """Args:
            data: The configuration dictionary.
            parent: the SiteConfig from which this data was extracted
                (this is stored as self.parent, but not used).

        The configuration dictionary should have the following elements:

        ``wamp_server`` (required): URL to the WAMP router's websocket
            access point for ocs.  E.g., ``ws://host-2:8001/ws``.
            WAMP routers can have multiple access points, with
            different protocols, security layers, and permissions.
            (Command line override: ``--site-hub``.)

        ``wamp_http`` (optional): URL to the WAMP router's http bridge
            interface.  This is the best interface for simple clients
            to use.  E.g., ``http://host-2:8001/call``.

        ``wamp_realm`` (required): The WAMP realm to use.  WAMP
            clients operating in a particular realm are isolated from
            clients connected to other realms.  Example and test code
            will often use ``debug_realm`` here.  (Command line
            override: ``--site-realm``.)

        ``address_root`` (required): The base address to be used by
            all OCS Agents.  This is normally something simple like
            ``observatory`` or ``detlab``.  (Command line override:
            ``--address-root``.)

        """
        self = cls()
        self.parent = parent
        self.data = data
        return self

    def summary(self):
        return summarize_dict(self.data)


class InstanceConfig:

    _MANAGE_MAP = {
        # Fundamental states
        'host/up': 'host/up',
        'host/down': 'host/down',
        'docker/up': 'docker/up',
        'docker/down': 'docker/down',
        'ignore': 'ignore',

        # Aliases for deprecated yes / no.
        'yes': 'host/up',
        'no': 'ignore',

        # Non-deprecated aliases.
        'docker': 'docker/up',
        'host': 'host/up',
        'up': 'host/up',
        'down': 'host/down',

        # Default.
        None: 'host/up',
    }

    def __init__(self):
        self.arguments = []

    @classmethod
    def from_dict(cls, data, parent=None):
        """Args:
            data: The configuration dictionary.
            parent: the HostConfig from which this data was extracted
                (this is stored as self.parent, but not used).

        The configuration dictionary should have the following elements:

        ``instance-id`` (str, required)
            This string is used to set the Agent instance's base
            address.  This may also be matched against the instance-id
            provided by the Agent instance, as a way of finding the
            right InstanceConfig.

        ``agent-class`` (str, optional)
            Name of the Agent class.  This
            may be matched against the agent_class name provided by
            the Agent instance, as a way of finding the right
            InstanceConfig.

        ``arguments`` (list, optional):
            A list of arguments that should be passed back to the
            agent.  Historically the arguments have been grouped into
            into key value pairs, e.g. [['--key1', 'value'],
            ['--key2', 'value']] but these days whatever you passed in
            gets flattened to a single list (i.e. that is equivalent
            to ['--key1', 'value', '--key2', 'value'].

        ``manage`` (str, optional):
            A string describing how a HostManager should manage this
            agent.  See notes.

        Notes:

            The ``manage`` value is only relevant if a HostManager is
            configured to operate on the host.  In that case, the
            HostManager's treatment of the agent instance depends on
            the value of ``manage``:

            - "ignore": HostManager will not attempt to manage the
              agent instance.
            - "host/up": HostManager will manage the agent instance,
              launching it on the host system.  On startup, the
              instance will be set to target_state "up" (i.e. the
              HostManager will try to start it).
            - "host/down": like host/up, but HostManager will not
              start up the agent instance until explicitly requested
              to do.
            - "docker/up": HostManager will manage the agent instance
              through Docker.  On Startup, the instance will be set to
              target_state "up".
            - "docker/down": Like docker/up, but the instance will be
              forced to target_state "down" on startup.

            In earlier versions of OCS, the acceptable values were
            "yes", "no", and "docker".  Those were equivalent to
            current values of "host/down", "ignore", and "docker/down".

            Those values are still accepted, but note that "yes" and
            "docker" are now equivalent to "host/up" and "docker/up".

            The following abbreviated values are also accepted:

            - "host": same as "host/up"
            - "up": same as "host/up"
            - "down": same as "host/down"

        """
        self = cls()
        self.parent = parent
        self.data = data
        self.arguments = self.data.get('arguments', [])
        self.manage = self.data.get('manage')
        self.manage = self._MANAGE_MAP.get(self.manage, self.manage)
        return self


def summarize_dict(d):
    output = '\n'.join(['  %s: %s,' % (repr(k), repr(v))
                        for k, v in d.items()])
    return '{\n%s\n}' % output


class ArgContainer:
    """
    A container to store a list of args as a dictionary, with the argument names
    (beginning with a hyphen) as keys, and list of arguments as values. Any
    arguments passed before an argument key is put under the '__positional__'
    key, even though positional arguments aren't really supported by ocs agents
    or the site-config....

    Args:
        args (list):
            Argument list (each item should be a single word)

    Attributes:
        arg_dict (dict):
            Dictionary of arguments, indexed by argument keyword.
    """

    def __init__(self, args):
        self.arg_dict = collections.OrderedDict()

        cur_key = '__positional__'
        self.arg_dict[cur_key] = []

        def is_new_arg(arg):
            if arg[0] != '-':
                return False
            # Check that first character after '-' is not a digit
            if not arg.strip('-')[0].isalpha():
                return False
            return True

        for arg in args:
            if is_new_arg(arg):
                cur_key = arg
                self.arg_dict[cur_key] = []
            else:
                self.arg_dict[cur_key].append(arg)

    def update(self, arg_container2):
        """
        Updates the arg_dict with the arg_dict from another ArgContainer

        Args:
            arg_container2 (ArgContainer):
                The other ArgContainer with which you want to update the arg_dict.
        """
        self.arg_dict.update(arg_container2.arg_dict)

    def to_list(self):
        """
        Returns the argument list representation of this container.
        """
        arg_list = []
        for k, v in self.arg_dict.items():
            if k != '__positional__':
                arg_list.append(k)
            arg_list.extend(v)

        return arg_list


def add_arguments(parser=None):
    """
    Add OCS site_config options to an ArgumentParser.

    Args:
        parser: an ArgumentParser.  If this is None, a new parser is
            created.

    Returns:
        The ArgumentParser that was passed in, or the new one.

    Arguments include the ``--site-*`` family.  See code or online
    documentation for details.
    """
    # Note that we use sphinxarg.ext to expose the help=... text in
    # the online sphinx docs.

    """
    ``--site=...``
        Instead of the default site, use the configuration
        for the specified site.  The configuration is loaded from
        ``$OCS_CONFIG_DIR/{site}.yaml``.  If --site=none, the
        site_config facility will not be used at all.

    ``--site-file=...``
        Instead of the default site config, use the
        specified file.  Full path must be specified.

    ``--site-host=...``
        Override the OCS determination of what host this instance is
        running on, and instead use the configuration for the
        indicated host.

    ``--site-hub=...``:
        Override the ocs hub url (wamp_server).

    ``--site-http=...``:
        Override the ocs hub http url (wamp_http).

    ``--site-realm=...``:
        Override the ocs hub realm (wamp_realm).

    ``--instance-id=...``:
        Look in the SCF for Agent-instance specific configuration
        options, and use those to launch the Agent.

    ``--address-root=...``:
        Override the site default address root.

    ``--log-dir=...``:
        Override the host default logging directory.

    ``--working-dir=...``:
        Propagate the working directory.

    ``--crossbar-timeout=...``:
        Length of time in seconds that the Agent will try to reconnect to the
        crossbar server before shutting down.

    """
    if parser is None:
        parser = argparse.ArgumentParser()
    group = parser.add_argument_group('Site Config Options')
    group.add_argument('--site', help="Instead of the default site, use the "
                       "configuration for the specified site.  The configuration is loaded "
                       "from ``$OCS_CONFIG_DIR/{site}.yaml``.  If ``--site=none``, the "
                       "site_config facility will not be used at all.")
    group.add_argument('--site-file', help="Instead of the default site config, "
                       "use the specified file. Full path must be specified.")
    group.add_argument('--site-host', help="Override the OCS determination of "
                       "what host this instance is running on, and instead use the "
                       "configuration for the indicated host.")
    group.add_argument('--site-hub', help="Override the ocs hub url (wamp_server).")
    group.add_argument('--site-http', help="Override the ocs hub http url (wamp_http).")
    group.add_argument('--site-realm', help="Override the ocs hub realm (wamp_realm).")
    group.add_argument('--instance-id', help="Look in the SCF for "
                       "Agent-instance specific configuration options, and use those to launch "
                       "the Agent.")
    group.add_argument('--address-root', help="Override the site default address root.")
    group.add_argument('--registry-address', help="Deprecated.")
    group.add_argument('--log-dir', help="Set the logging directory.")
    group.add_argument('--working-dir', help="Propagate the working directory.")
    group.add_argument('--crossbar-timeout', type=int, help="Length of time in seconds "
                       "that the Agent will try to reconnect to the crossbar server before "
                       "shutting down. Note this is set per Agent in an instance's arguments list.")
    return parser


def get_config(args, agent_class=None):
    """
    Args:
        args: The argument object returned by
            ArgumentParser.parse_args(), or equivalent.  It is assumed
            that all properties defined by "add_arguments" are present
            in this object.
        agent_class: Class name passed in to match against the list of
            device classes in each host's list.

    Special values accepted for agent_class:
    - '*control*': do not insist on matching host or device.
    - '*host*': do not insist on matching device (but do match host).

    Returns:
        The tuple (site_config, host_config, device_config).
    """
    if args.site == 'none':
        return (None, None, None)

    site_file = args.site_file
    site = args.site
    if site_file is None:
        if site is None:
            site = 'default'
        assert (os.getenv('OCS_CONFIG_DIR') is not None)
        site_file = os.path.join(os.getenv('OCS_CONFIG_DIR'),
                                 site + '.yaml')
    else:
        assert (site is None)  # do not pass both --site and --site-file

    # Load the site config file.
    site_config = SiteConfig.from_yaml(site_file)

    # Matching behavior.
    no_host_match = (agent_class == '*control*')
    no_dev_match = no_host_match or (agent_class == '*host*')

    # Identify our host and update site.hub.
    host_config = None
    if args.site_host is not None:
        host_attempts = [args.site_host, 'localhost']
    else:
        host_attempts = [socket.gethostname(), 'localhost']
    for host_try in host_attempts:
        if host_try in site_config.hosts:
            host_config = site_config.hosts[host_try]
            host_update_dict = {
                k: host_config.data[k]
                for k in ['wamp_server', 'wamp_http', 'wamp_realm']
                if k in host_config.data.keys()
            }
            site_config.hub.data.update(host_update_dict)
            # Updates host_config with command line args
            if args.working_dir is not None:
                host_config.working_dir = args.working_dir
            if args.log_dir is not None:
                host_config.log_dir = args.log_dir
            break
    else:
        if not no_host_match:
            raise KeyError('Site config has no entry in "hosts" for {}'
                           .format(host_attempts))

    # Override the WAMP hub?
    if args.site_hub is not None:
        site_config.hub.data['wamp_server'] = args.site_hub

    if args.site_http is not None:
        site_config.hub.data['wamp_http'] = args.site_http

    # Override the realm?
    if args.site_realm is not None:
        site_config.hub.data['wamp_realm'] = args.site_realm

    # Identify our agent-instance.
    instance_config = None
    if no_dev_match:
        pass
    elif args.instance_id is not None:
        # Find the config for this instance-id.
        for dev in host_config.instances:
            if dev['instance-id'] == args.instance_id:
                instance_config = InstanceConfig.from_dict(
                    dev, parent=host_config)
                break
    else:
        # Use the agent_class to figure it out...
        for dev in host_config.instances:
            if dev['agent-class'] == agent_class:
                if instance_config is not None:
                    raise RuntimeError(
                        f"Multiple matches found for agent-class={agent_class}"
                        " ... you probably need to pass --instance-id=")
                instance_config = InstanceConfig.from_dict(
                    dev, parent=host_config)
    if instance_config is None and not no_dev_match:
        raise RuntimeError("Could not find matching device description.")
    return collections.namedtuple('SiteConfig', ['site', 'host', 'instance'])(site_config, host_config, instance_config)


def add_site_attributes(args, site, host=None):
    """
    Adds site and host attributes to namespace if they do not exist.

    Args:
        args:
            namespace to add attributes to.
        site:
            Site config object.
        host:
            Host config object.
    """
    if args.site_hub is None:
        args.site_hub = site.hub.data['wamp_server']
    if args.site_http is None:
        args.site_http = site.hub.data.get('wamp_http')
    if args.site_realm is None:
        args.site_realm = site.hub.data['wamp_realm']
    if args.address_root is None:
        args.address_root = site.hub.data['address_root']
    if (args.log_dir is None) and (host is not None):
        args.log_dir = host.log_dir
    if (args.crossbar_timeout is None) and (host is not None):
        args.crossbar_timeout = host.crossbar_timeout


@deprecation.deprecated(deprecated_in='v0.6.0',
                        details="Use site_config.parse_args instead")
def reparse_args(args, agent_class=None):
    """
    THIS FUNCTION IS NOW DEPRECATED... Use the parse_args function instead
    to parse command line and site-config args simultaneously.

    Process the site-config arguments, and modify them in place
    according to the agent-instance's computed instance-id.

    Args:
        args: The argument object returned by
            ArgumentParser.parse_args(), or equivalent.

        agent_class: Class name passed in to match against the list of
            device classes in each host's list.

    Special values accepted for agent_class:
    - '*control*': do not insist on matching host or device.
    """
    if args.site == 'none':
        return args

    site, host, instance = get_config(args, agent_class=agent_class)

    add_site_attributes(args, site, host=host)

    if instance is not None:
        if args.instance_id is None:
            args.instance_id = instance.data['instance-id']

        for k, v in instance.data['arguments']:
            kprop = k.lstrip('-').replace('-', '_')
            print('site_config is setting values of "%s" to "%s".' % (kprop, v))
            setattr(args, kprop, v)

    return args


def get_control_client(instance_id, site=None, args=None, start=True,
                       client_type='http'):
    """Instantiate and return a client_http.ControlClient, targeting the
    specified instance_id.

    Args:
        site (SiteConfig): All configuration will be taken from this
            object, if it is not None.

        args: Arguments from which to derive the site configuration.
            If this is None, then the arguments from the command line
            are parsed through the usual site_config system.  If this
            is a list of strings, then these arguments will be parsed
            instead of sys.argv[1:].  Note that to use the default
            configuration (without looking at sys.argv), pass args=[].
            It is also permitted to pass a pre-parsed
            argparse.Namespace object (or similar).

        start (bool): Determines whether to call .start() on the client before
            returning it.

        client_type (str): Select the client type, currently only 'http'.
            wamp_http address must be known. Note that 'wampy' used to be a
            supported type, but was dropped in OCS v0.8.0.

    Returns a ControlClient.

    """
    if site is None:
        if args is None:
            args = sys.argv[1:]
        if not hasattr(args, 'instance_id'):
            # If it doesn't have .instance_id, it's not a parsed
            # Namespace so let's assume it's a list of strings.
            args = ocs.site_config.parse_args(agent_class='*control*',
                                              args=args)
        site, _, _ = ocs.site_config.get_config(args, '*control*')
    full_addr = '%s.%s' % (site.hub.data['address_root'], instance_id)
    if client_type is None:
        if site.hub.data.get('wamp_http'):
            client_type = 'http'
        else:
            client_type = 'wampy'
    if client_type == 'wampy':
        raise ValueError('client_type %s no longer supported' % client_type)
    elif client_type == 'http':
        from ocs import client_http
        client = client_http.ControlClient(
            full_addr,
            url=site.hub.data['wamp_http'],
            realm=site.hub.data['wamp_realm'])
    else:
        raise ValueError('Unknown client_type request: %s' % client_type)
    return client


# We'll also keep the Agent script registry here.
agent_script_reg = {}


def register_agent_class(class_name, filename):
    """Register an Agent script in the site_config registry.

    Args:
        class_name (str): The Agent class name, e.g. "HostManager".
        filename (str): The full path to the script that launches an
            instance of this Agent class.

    """
    agent_script_reg[class_name] = filename


def scan_for_agents(do_registration=True):
    """Identify and import ocs Agent plugin scripts.  This will find all
    modules in the current module search path (sys.path) that begin
    with the name 'ocs_plugin\\_'.

    Args:
        do_registration (bool): If True, the modules are imported,
            which likely causes them to call register_agent_class on
            each agent they represent.

    Returns:
        The list of discovered module names.

    """
    import pkgutil
    import importlib
    items = []
    for modinfo in pkgutil.iter_modules():
        if modinfo.name.startswith('ocs_plugin_'):
            items.append(modinfo.name)
            if do_registration:
                importlib.import_module(modinfo.name)
    return items


def parse_args(agent_class=None, parser=None, args=None):
    """
    Function to parse site-config and agent arguments. This function takes
    site, host, and instance arguments into account by making sure the instance
    arguments get passed through the arg_parse parser. This helps make sure
    units and options are consistent with those defined by the argparse
    argument, even when the arguments come from the site-config file and not
    the command line.

    Args:
        agent_class (str, optional):
            Name of the Agent class.  This
            may be matched against the agent_class name provided by
            the Agent instance, as a way of finding the right
            InstanceConfig.
        parser (argparse.ArgumentParser, optional):
            Argument parser containing agent-specific arguments.
            If None, an empty parser will be created.
        args (list of str):
            Arguments to parse; defaults to sys.argv[1:].

    Returns:
        An argparse.Namespace, as you would get from
        parser.parse_args().

    """

    # Creates pre_parser
    pre_parser = argparse.ArgumentParser()
    add_arguments(pre_parser)

    # Full parser
    if parser is None:
        parser = argparse.ArgumentParser()
    add_arguments(parser)

    if args is None:
        args = sys.argv[1:]

    # Intercepts help commands to print full usage statement
    if any(h in args for h in ['-h', '--help']):
        # Instead of print_help(), trust parse_args() to run resolve
        # to any sub-parsers and print context-appropriate help.
        parser.parse_args(args=args)
        parser.exit()  # shouldn't get here.

    pre_args, _ = pre_parser.parse_known_args(args=args)

    site, host, instance = get_config(pre_args, agent_class=agent_class)

    if instance is not None:
        # When the user omits instance_id, it can still be matched,
        # through agent_class by the get_config parser.  In that case,
        # copy its value into args.
        if pre_args.instance_id is None:
            instance.arguments.append(['--instance-id',
                                       instance.data['instance-id']])

    # Container from command line args
    cl_container = ArgContainer(args)

    # Flattens instance arguments to single non-nested list
    def flatten(container):
        out = []
        for i in container:
            if isinstance(i, (list, tuple)):
                out.extend(flatten(i))
            else:
                out.append(i)
        return out

    arg_container = ArgContainer([])

    if instance is not None:
        arg_container.update(ArgContainer(map(str, flatten(instance.arguments))))

    # Replace site values with command line values if they exist
    arg_container.update(cl_container)

    # Parse combined CL + site arguments
    args = parser.parse_args(args=arg_container.to_list())

    # Add site and host attributes to args namespace
    add_site_attributes(args, site, host=host)

    # Add agent_class attribute.
    if not hasattr(args, 'agent_class'):
        setattr(args, 'agent_class', agent_class)

    return args



#!/usr/bin/env python3

import os
import stat
import sys
from ocs import site_config

"""The idea here is to create a launcher script that will live in the
OCS_CONFIG_DIR directory, and also a minimal systemd script that gets
installed to the system.

The launcher script is helpful for configurability -- for example, to
make use of a non-system Python.  If we accept the usefulness/
necessity of the launcher script, then there is not much reason to
create other communication channels between the systemd service file
and the agent.  (For example, one could imagine setting lots of
environment variables in the service, and having host_manager.py fully
self-configure from those.)

"""

SERVICE_DEFAULT = '/etc/systemd/system/'

SYSTEMD_TEMPLATE = """[Unit]
Description=OCS HostManager{host_detail}

[Service]
ExecStart={cmd}
User={service_user}
Restart=always
RestartSec=10s
{environment_lines}

[Install]
WantedBy=multi-user.target
"""

LAUNCHER_TEMPLATE = """#!{shell}

# This launcher script is for use by systemd; a corresponding systemd
# service file was probably installed to {systemd_dest}.

### Add / modify environment and launch variables here.

SITE_FILE={site_file}
PYTHON={python_bin}

###

${{PYTHON}} -m ocs.agent_cli \\
  --site-file=${{SITE_FILE}} \\
  {host_manager_args}
"""


def get_parser():
    import argparse
    parser = argparse.ArgumentParser()
    group = parser.add_argument_group('Options specific to systemd service creation')
    group.add_argument('--docker-compose', action='append', default=[])
    # group.add_argument('--host', help=)
    group.add_argument('--shell', help="Override the shell used for the launcher script.")
    group.add_argument('--python-bin', help="Select the Python interpreter with which to launch HostManager.")
    group.add_argument('--launcher-dir', help="Override the directory of the launcher script.")
    group.add_argument('--launcher-script', help="Override the name of the launcher script.")
    group.add_argument('--service-dir', help="Directory to which the .service file will be written; "
                       f"default is {SERVICE_DEFAULT}.")
    group.add_argument('--service-user', help="Override the user under which HostManager agent will be run.")
    group.add_argument('--service-name', help="Override the name of the systemd service.")
    group.add_argument('--service-host', help="Set a hostname to use when generating the service name.")
    return parser


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = get_parser()
    args = site_config.parse_args(agent_class='*host*',
                                  parser=parser)
    site, host, _ = site_config.get_config(args, agent_class='*host*')
    assert (host.name is not None)  # It won't be, right?

    # The "hostname" is potentially used for a few things:
    # - to name the service
    # - to name the launcher script
    # - to configure the host-manager
    #
    # It's only sensible to have the launcher script name and host-manager
    # use the same value.  For naming the service, the sensible default is
    # to not include the host name, since a most natural configuration
    # would have just one host-manager on the host.
    #
    # So ... site_config host.name will be used for the
    # launcher/host-manager.  And the service name will default to a
    # hostless version unless args.service_host is passed.

    if args.service_name is None:
        if args.service_host is None:
            args.service_name = 'ocs-hostmanager.service'
        else:
            args.service_name = f'ocs-hostmanager-{args.service_host}.service'
    if args.instance_id is None:
        args.instance_id = f'hm-{host.name}'
    if args.shell is None:
        args.shell = os.environ['SHELL']
    if args.launcher_dir is None:
        args.launcher_dir = os.environ['OCS_CONFIG_DIR']
    if args.launcher_script is None:
        args.launcher_script = os.path.join(args.launcher_dir, f'launcher-{args.instance_id}.sh')
    args.launcher_script = os.path.abspath(args.launcher_script)
    if args.service_user is None:
        args.service_user = os.environ['USER']
    if args.service_dir is None:
        args.service_dir = SERVICE_DEFAULT
    if args.service_name is None:
        args.service_name = f'ocs-{args.instance_id}.service'
    args.systemd_dest = os.path.join(args.service_dir, args.service_name)

    if args.python_bin is None:
        args.python_bin = sys.executable
    if args.site_file is None:
        args.site_file = site.source_file

    host_manager_args = [
        f'--instance-id={args.instance_id}',
        f'--site-host={host.name}',
    ]
    if len(args.docker_compose):
        args.docker_compose = [os.path.abspath(f) for f in args.docker_compose]
        host_manager_args.append('--docker-compose=' + ','.join(args.docker_compose))

    host_detail = ' for %s' % host.name
    systemd_script = SYSTEMD_TEMPLATE.format(
        cmd=args.launcher_script,
        host_detail=host_detail,
        environment_lines='',
        **vars(args))

    launcher_code = LAUNCHER_TEMPLATE.format(
        host_manager_args='  \\\n  '.join(host_manager_args),
        **vars(args))

    print(f'Writing {args.launcher_script} ...')
    open(args.launcher_script, 'w').write(launcher_code)
    os.chmod(args.launcher_script, os.stat(args.launcher_script).st_mode
             | (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH))

    print(f'Writing {args.systemd_dest} ...')
    try:
        open(args.systemd_dest, 'w').write(systemd_script)
    except PermissionError:
        print()
        print(f' -- PermissionError trying to write to {args.systemd_dest}!')
        print()
        print('Re-run the previous command, but with --service-dir=./ to create\n'
              'the service file in the current directory.  Then install it with:')
        print()
        print(f'   sudo cp ./{args.service_name} {args.service_dir}')
        print('   sudo systemctl daemon-reload')
